import math

import torch
from torch.utils.data import Sampler, RandomSampler, BatchSampler

__all__ = [
    'TBPTTSampler',
    'IndicesSampler'
]


class TBPTTSampler(Sampler):
    """
    yields batches of indices for performing Truncated Back Propagation Through Time
    """

    def __init__(self,
                 n_samples,
                 batch_size=64,  # nbr of "tracks" per batch
                 chunk_length=8 * 16000,  # total length of a track
                 seq_len=512,  # nbr of samples per backward pass
                 oversampling=1,
                 ):
        super().__init__(None)
        self.n_samples = n_samples
        self.chunk_length = min(chunk_length, n_samples)
        self.seq_len = seq_len
        self.n_chunks = max(1, self.n_samples // self.chunk_length - int(oversampling > 1))
        self.remainder = self.n_samples % self.chunk_length
        self.n_per_chunk = self.chunk_length // self.seq_len
        self.batch_size = batch_size
        self.oversampling = oversampling

    def __iter__(self):
        indices = torch.arange(self.n_chunks * self.oversampling)
        smp = RandomSampler(indices)
        for top in BatchSampler(smp, self.batch_size, False):
            # for _ in range(self.oversampling):
            offsets = torch.randint(0, self.remainder, (self.batch_size,))
            top_idx = tuple(o + ((t % self.n_chunks) * self.chunk_length)
                            for t, o in zip(top, offsets))
            for start in range(self.n_per_chunk):
                # start indices of the batch
                yield tuple(t + (start * self.seq_len) for t in top_idx)

    def __len__(self):
        return (self.oversampling * self.n_chunks // self.batch_size) * self.n_per_chunk


class IndicesSampler(Sampler):
    def __init__(self,
                 N=0,
                 indices=(),
                 min_i=0,
                 max_i=None,
                 redraw=True,
                 ):
        super().__init__(None)
        self.N = N
        self._indices = indices
        self.min_i = min_i
        self.max_i = max_i
        self.redraw = redraw
        self.indices = self.draw_indices(N, indices)

    def __iter__(self):
        for i in self.indices:
            yield i
        if self.redraw:
            self.indices = self.draw_indices(self.N, self._indices)

    def draw_indices(self, N, indices):
        n_idx = len(indices)
        if N == 0 and n_idx == 0:
            raise ValueError("`indices` can not be empty if `N` == 0")
        elif N > 0 and n_idx == 0:
            # only random
            return torch.randint(self.min_i, self.max_i, (N,))
        elif N == 0 and n_idx > 0:
            # only static
            return indices
        elif N > 0 and 0 < n_idx < N:
            # complete statics with randoms
            return [*indices, *torch.randint(self.min_i, self.max_i, (N,))]
        else:
            # N <= n_idx -> only statics
            return indices



import h5mapper as h5m
import torch
import torch.nn as nn

from .io_modules import *
from ..features import Spectrogram, MuLawSignal
from ..networks import Seq2SeqLSTMNetwork as S2SNet, MultiSeq2SeqLSTM
from ..modules import Flatten


__all__ = [
    "Seq2SeqLSTMv0",
    "Seq2SeqLSTM",
    "Seq2SeqMuLaw"
]


class Seq2SeqLSTMv0(S2SNet):
    feature = Spectrogram(sr=22050, n_fft=2048, hop_length=512, coordinate='mag')
    """no input module"""
    def __init__(self,
                 feature=None,
                 output_heads=4,
                 scaled_activation=True,
                 phs="a",
                 **net_hp):
        model_dim = net_hp['model_dim']
        fft_dim = (feature.n_fft // 2 + 1)
        if feature.coordinate == "mag":
            _, outpt_mod = mag_spec_io(fft_dim, model_dim, 1, output_heads, scaled_activation)
        elif feature.coordinate == "pol":
            _, outpt_mod = pol_spec_io(fft_dim//2, model_dim, 1, output_heads, scaled_activation, phs)
            net_hp["input_module"] = Flatten(2)
        else:
            raise ValueError(f"Seq2SeqLSTM doesn't support coordinate of type {feature.coordinate}")
        net_hp["input_dim"] = fft_dim
        net_hp["output_module"] = outpt_mod
        super(Seq2SeqLSTMv0, self).__init__(**net_hp)
        self.hp.feature = self.feature = feature
        self.hp.output_heads = output_heads
        self.hp.scaled_activation = scaled_activation
        self.hp.phs = phs


class Seq2SeqLSTM(S2SNet):
    feature = Spectrogram(sr=22050, n_fft=2048, hop_length=512, coordinate='mag')

    def __init__(self,
                 feature=None,
                 input_heads=1,
                 output_heads=4,
                 scaled_activation=True,
                 phs="a",
                 **net_hp):
        fft_dim = feature.n_fft // 2 + 1
        model_dim = net_hp['model_dim']
        if feature.coordinate == "mag":
            inpt_mod, outpt_mod = mag_spec_io(fft_dim, model_dim, input_heads, output_heads, scaled_activation)
        elif feature.coordinate == "pol":
            inpt_mod, outpt_mod = pol_spec_io(fft_dim, model_dim, input_heads, output_heads, scaled_activation, phs)
        else:
            raise ValueError(f"Seq2SeqLSTM doesn't support coordinate of type {feature.coordinate}")
        net_hp["input_dim"] = model_dim
        net_hp["input_module"] = inpt_mod
        net_hp["output_module"] = outpt_mod
        super(Seq2SeqLSTM, self).__init__(**net_hp)
        self.hp.feature = self.feature = feature
        self.hp.input_heads = input_heads
        self.hp.output_heads = output_heads
        self.hp.scaled_activation = scaled_activation
        self.hp.phs = phs


class Seq2SeqMuLaw(Seq2SeqLSTM):
    feature = MuLawSignal(sr=16000, q_levels=256, normalize=True)

    def __init__(self, feature=None, mlp_dim=128, **net_hp):
        net_hp["input_module"] = feature.input_module(net_hp["input_dim"])
        out_d = net_hp["model_dim"]
        net_hp["output_module"] = feature.output_module(out_d, mlp_dim=mlp_dim)
        super(Seq2SeqMuLaw, self).__init__(**net_hp)
        self.hp.feature = self.feature = feature
        self.hp.mlp_dim = mlp_dim


class MultiSeq2SeqFFT(MultiSeq2SeqLSTM):
    pass


class MultiSeq2SeqMuLaw(MultiSeq2SeqLSTM):
    pass


class NoNANNEt(nn.Module):

    def attach_hooks(self):

        def frw_hook(module, inpt, output):
            is_tensor = lambda x: isinstance(x, torch.Tensor)
            h5m.process_batch(inpt, is_tensor,
                              lambda x: print("INPUT NAN", module) if torch.any(torch.isnan(x)) else None
                              )
            h5m.process_batch(output, is_tensor,
                              lambda x: print("OUTPUT NAN", module) if torch.any(torch.isnan(x)) else None
                              )
            return output

        def grad_hook(module, inpt, output):
            is_tensor = lambda x: isinstance(x, torch.Tensor)
            h5m.process_batch(inpt, is_tensor,
                              lambda x: print("INPUT GRAD NAN", module) if torch.any(torch.isnan(x)) else None
                              )
            h5m.process_batch(output, is_tensor,
                              lambda x: print("OUTPUT GRAD NAN", module) if torch.any(torch.isnan(x)) else None
                              )
            return inpt

        for mod in self.modules():
            mod.register_full_backward_hook(grad_hook)
            mod.register_forward_hook(frw_hook)

        def hook(grad):
            is_nans = torch.isnan(grad)
            if torch.any(is_nans):
                return torch.where(is_nans, torch.zeros_like(grad), grad)
            return grad
        for p in self.parameters():
            p.register_hook(hook)


import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from functools import reduce
import operator as op


class ConvTranspose1dWaveGanGen(nn.Module):
    def __init__(
            self,
            in_channels,
            out_channels,
            kernel_size,
            up_sample=4,
            up_sampling="conv",
            use_batch_norm=False,
    ):
        super(ConvTranspose1dWaveGanGen, self).__init__()
        self.up_sample = up_sample
        self.up_sampling = up_sampling

        if up_sampling == "nearest":
            reflection_pad = nn.ReflectionPad1d(
                (kernel_size // 2) if kernel_size % 2 == 1 else (kernel_size // 2 - 1, kernel_size // 2)
            )
            conv1d = nn.Conv1d(in_channels, out_channels, kernel_size, stride=(1,))
            ops = [reflection_pad, conv1d]
        else:
            Conv1dTrans = nn.ConvTranspose1d(
                in_channels, out_channels, kernel_size, stride=(up_sample,),
                padding=(math.ceil((kernel_size - up_sample) / 2) if kernel_size > up_sample else 0,),
                output_padding=(up_sample - kernel_size if kernel_size < up_sample else int(kernel_size % 2 == 1),)
            )
            ops = [Conv1dTrans]

        if use_batch_norm:
            ops.append(nn.BatchNorm1d(out_channels))

        self.ops = nn.Sequential(*ops)

    def forward(self, x):
        if self.up_sampling == "nearest":
            # recommended by wavgan paper to use nearest up_sampling
            x = nn.functional.interpolate(x, scale_factor=self.up_sample, mode="nearest")
        return self.ops(x)


class WaveGANGenerator(nn.Module):
    def __init__(
            self,
            latent_dim=100,
            model_size=64,
            kernel_size=25,
            t0=16,
            up_sample=(4, ),
            up_sampling="conv",  # or "nearest"
            n_layers=5,
            n_channels=1,
            post_proc_filt_len=512,
            use_batch_norm=False,
            verbose=False,
    ):
        super(WaveGANGenerator, self).__init__()
        self.latent_dim = latent_dim
        self.model_size = model_size  # d
        self.kernel_size = kernel_size
        self.t0 = t0
        self.n_channels = n_channels  # c
        self.post_proc_filt_len = post_proc_filt_len
        self.use_batch_norm = use_batch_norm
        self.verbose = verbose
        self.dim_mul = dim_mul = reduce(op.mul, up_sample)
        self.fc1 = nn.Linear(latent_dim, model_size * dim_mul * t0)
        self.bn1 = nn.BatchNorm1d(num_features=model_size * dim_mul) if use_batch_norm else None

        deconv_layers = [
            ConvTranspose1dWaveGanGen(
                (dim_mul * model_size) // (2 ** downscaling_factor),
                (dim_mul * model_size) // (2 ** (downscaling_factor + 1)),
                kernel_size,
                up_sample=up,
                up_sampling=up_sampling,
                use_batch_norm=use_batch_norm,
            ) for up, downscaling_factor in zip(up_sample, range(n_layers - 1))] + \
            [
                ConvTranspose1dWaveGanGen(
                    (dim_mul * model_size) // (2 ** (n_layers - 1)),
                    n_channels,
                    kernel_size,
                    up_sample=up_sample[-1],
                    up_sampling=up_sampling,
                )
            ]

        self.deconv_list = nn.ModuleList(deconv_layers)
        self.ppfilter1 = nn.Conv1d(n_channels, n_channels, (post_proc_filt_len,)) if post_proc_filt_len else None

        for m in self.modules():
            if isinstance(m, nn.ConvTranspose1d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight.data)

    def forward(self, x):
        x = self.fc1(x).view(-1, self.dim_mul * self.model_size, self.t0)
        if self.use_batch_norm:
            x = self.bn1(x)
        x = F.relu(x)
        if self.verbose:
            print(self.fc1.__class__.__qualname__, x.shape)

        for deconv in self.deconv_list[:-1]:
            x = F.relu(deconv(x))
            if self.verbose:
                print(deconv.__class__.__qualname__, x.shape)
        if self.ppfilter1 is None:
            output = torch.tanh(self.deconv_list[-1](x))
        else:
            x = F.relu(self.deconv_list[-1](x))
            # Pad for "same" filtering
            if (self.post_proc_filt_len % 2) == 0:
                pad_left = self.post_proc_filt_len // 2
                pad_right = pad_left - 1
            else:
                pad_left = (self.post_proc_filt_len - 1) // 2
                pad_right = pad_left
            x = F.pad(x, (pad_left, pad_right))
            output = torch.tanh(self.ppfilter1(x))
        return output


class PhaseShuffle(nn.Module):
    """
    Performs phase shuffling, i.e. shifting feature axis of a 3D tensor
    by a random integer in {-n, n} and performing reflection padding where
    necessary.
    """

    # Copied from https://github.com/jtcramer/wavegan/blob/master/wavegan.py#L8
    def __init__(self, shift_factor):
        super(PhaseShuffle, self).__init__()
        self.shift_factor = shift_factor

    def forward(self, x):
        if self.shift_factor == 0:
            return x
        # uniform in (L, R)
        k_list = (
                torch.Tensor(x.shape[0]).random_(0, 2 * self.shift_factor + 1)
                - self.shift_factor
        )
        k_list = k_list.numpy().astype(int)

        # Combine sample indices into lists so that less shuffle operations
        # need to be performed
        k_map = {}
        for idx, k in enumerate(k_list):
            k = int(k)
            if k not in k_map:
                k_map[k] = []
            k_map[k].append(idx)

        # Make a copy of x for our output
        x_shuffle = x.clone()

        # Apply shuffle to each sample
        for k, idxs in k_map.items():
            if k > 0:
                x_shuffle[idxs] = F.pad(x[idxs][..., :-k], (k, 0), mode="reflect")
            else:
                x_shuffle[idxs] = F.pad(x[idxs][..., -k:], (0, -k), mode="reflect")

        assert x_shuffle.shape == x.shape, "{}, {}".format(x_shuffle.shape, x.shape)
        return x_shuffle


class Conv1DWaveGanDisc(nn.Module):
    def __init__(
            self,
            input_channels,
            output_channels,
            kernel_size,
            down_sample=4,
            alpha=0.2,
            shift_factor=2,

            use_batch_norm=False,
            drop_prob=0,
    ):
        super(Conv1DWaveGanDisc, self).__init__()
        self.conv1d = nn.Conv1d(
            input_channels, output_channels, kernel_size, stride=(down_sample,),
            padding=(math.ceil((kernel_size - down_sample) / 2) if kernel_size > down_sample else 0,)
        )
        self.batch_norm = nn.BatchNorm1d(output_channels)
        self.phase_shuffle = PhaseShuffle(shift_factor)
        self.alpha = alpha
        self.use_batch_norm = use_batch_norm
        self.use_phase_shuffle = shift_factor == 0
        self.use_drop = drop_prob > 0
        self.dropout = nn.Dropout2d(drop_prob)

    def forward(self, x):
        x = self.conv1d(x)
        if self.use_batch_norm:
            x = self.batch_norm(x)
        x = F.leaky_relu(x, negative_slope=self.alpha)
        if self.use_phase_shuffle:
            x = self.phase_shuffle(x)
        if self.use_drop:
            x = self.dropout(x)
        return x


class WaveGANDiscriminator(nn.Module):
    def __init__(
            self,
            model_size=64,
            kernel_size=25,
            down_sample=(4, ),
            n_layers=5,
            n_channels=1,
            shift_factor=2,
            alpha=0.2,
            verbose=False,
            use_batch_norm=False,
    ):
        super(WaveGANDiscriminator, self).__init__()

        self.model_size = model_size  # d
        self.use_batch_norm = use_batch_norm
        self.n_channels = n_channels  # c
        self.shift_factor = shift_factor  # n
        self.alpha = alpha
        self.verbose = verbose

        conv_layers = [
            Conv1DWaveGanDisc(
                n_channels,
                model_size,
                kernel_size,
                down_sample=down_sample[0],
                use_batch_norm=use_batch_norm,
                alpha=alpha,
                shift_factor=shift_factor,
            )
            ] + [
            Conv1DWaveGanDisc(
                model_size * (2 ** up_scaling_factor),
                model_size * (2 ** (up_scaling_factor + 1)),
                kernel_size,
                down_sample=down,
                use_batch_norm=use_batch_norm,
                alpha=alpha,
                shift_factor=shift_factor,
            )
            for down, up_scaling_factor in zip(down_sample, range(n_layers-1))]
        # self.fc_input_size = (down_sample ** (n_layers-1)) * model_size
        self.fc_input_size = reduce(op.mul, down_sample)

        self.conv_layers = nn.ModuleList(conv_layers)

        self.fc1 = nn.Linear(self.fc_input_size, 1)

        for m in self.modules():
            if isinstance(m, nn.Conv1d) or isinstance(m, nn.Linear):
                nn.init.kaiming_normal_(m.weight.data)

    def forward(self, x):
        for conv in self.conv_layers:
            x = conv(x)
            if self.verbose:
                print(conv.__class__.__qualname__, x.shape)
        x = x.view(-1, self.fc_input_size)
        if self.verbose:
            print(self.fc1.__class__.__qualname__, x.shape)

        return self.fc1(x)



import torch.nn as nn
import torch
from torch.nn import Transformer, TransformerEncoder, TransformerEncoderLayer
import math

__all__ = [
    'PositionalEncoding',
    'SimpleTransformer'
]


class PositionalEncoding(nn.Module):
    r"""Inject some information about the relative or absolute position of the tokens
        in the sequence. The positional encodings have the same dimension as
        the embeddings, so that the two can be summed. Here, we use sine and cosine
        functions of different frequencies.
    .. math::
        \text{PosEncoder}(pos, 2i) = sin(pos/10000^(2i/d_model))
        \text{PosEncoder}(pos, 2i+1) = cos(pos/10000^(2i/d_model))
        \text{where pos is the word position and i is the embed idx)
    Args:
        d_model: the embed dim (required).
        dropout: the dropout value (default=0.1).
        max_len: the max. length of the incoming sequence (default=5000).
    Examples:
        >>> pos_encoder = PositionalEncoding(d_model)
    """

    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)

    def forward(self, x):
        r"""Inputs of forward function
        Args:
            x: the sequence fed to the positional encoder model (required).
        Shape:
            x: [sequence length, batch size, embed dim]
            output: [sequence length, batch size, embed dim]
        Examples:
            >>> output = pos_encoder(x)
        """

        x = x + self.pe[:x.size(0), :]
        return self.dropout(x)


class SimpleTransformer(nn.Module):

    def _generate_square_subsequent_mask(self, sz):
        mask = (torch.triu(torch.ones(sz, sz)) == 1).transpose(0, 1)
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask

    def _generate_square_context_mask(self, sz, k=8):
        mask = torch.zeros(sz, sz)
        rg = torch.arange(k)
        for i in range(sz):
            mask[i, torch.clamp(rg + i - (2 // k), min=0, max=sz - 1)] = 1.
        mask = mask.float().masked_fill(mask == 0, float('-inf')).masked_fill(mask == 1, float(0.0))
        return mask

    def __init__(self, d_model, nhead, dim_feedforward, num_layers):
        super(SimpleTransformer, self).__init__()
        layer = TransformerEncoderLayer(d_model=d_model,
                                        nhead=nhead,
                                        dim_feedforward=dim_feedforward,
                                        dropout=0.0,
                                        activation="gelu")
        norm = nn.LayerNorm(d_model)
        self.model = TransformerEncoder(layer, num_layers=num_layers, norm=None)
        self.src_mask = None

    def forward(self, src):
        src = src.permute(1, 0, 2).contiguous()
        device = src.device
        if self.src_mask is None or self.src_mask.size(0) != len(src):
            mask = self._generate_square_context_mask(len(src)).to(device)
            self.src_mask = mask
        return self.model(src, self.src_mask).permute(1, 0, 2).contiguous()


from ..networks import SampleRNN as SampleRNNNet

__all__ = [
    "SampleRNN"
]


class SampleRNN(SampleRNNNet):

    def __init__(self, feature, chunk_length=16000 * 8, **net_hp):
        super(SampleRNN, self).__init__(**net_hp)
        self.feature = self.hp.feature = feature
        self.hp.chunk_length = chunk_length



import dataclasses as dtc

__all__ = [
    'Feature'
]


@dtc.dataclass
class Feature:
    """
    class for centralizing methods pertaining to a particular type, class or format of data.
    """
    __ext__ = None

    @property
    def params(self):
        """
        Returns
        -------
        params: dict
            the parameters of this Feature
        """
        return dtc.asdict(self)

    base_feature = None
    transform = None
    inverse_transform = None

    def input_module(self, net_dim):
        pass

    def output_module(self, net_dim):
        pass

    def batch_item(self, *args, **kwargs):
        pass

    def loss_fn(self, output, target):
        pass

    def post_create(self, db, schema_key):
        # since this is optional, we don't raise any exception
        pass

    def load(self, path):
        y = self.base_feature.load(path)
        return self.transform(y)

    def display(self, inputs, **waveplot_kwargs):
        y = self.inverse_transform(inputs)
        return self.base_feature.display(y, **waveplot_kwargs)

    def play(self, inputs):
        y = self.inverse_transform(inputs)
        return self.base_feature.play(y)

    def write(self, filename, inputs):
        y = self.inverse_transform(inputs)
        return self.base_feature.write(filename, y)


def feature(name,
            params={},
            base_feature=None,
            transform=None,
            inverse_transform=None,
            input_module=None,
            output_module=None,
            loss_fn=None):
    pass



from .io_modules import *
from .srnns import *
from .wavenets import *
from .s2s import *
from .nnn import *
from .checkpoint import *
from .ensemble import *


__all__ = [_ for _ in dir() if not _.startswith("_")]



import torch
import numpy as np
import muspy
import dataclasses as dtc
from typing import Optional
from collections import OrderedDict
import IPython.display as ipd

__all__ = [
    'MidiMusic',
    'MidiEvent'
]


class MMKNote(muspy.Note):
    """
    muspy.Note with `channel` and `program` attributes
    """
    _attributes = OrderedDict(
        [
            ("time", int),
            ("pitch", int),
            ("duration", int),
            ("velocity", int),
            ("pitch_str", str),
            ("channel", int),
            ("program", int)
        ]
    )
    _optional_attributes = ["velocity", "pitch_str", "channel", "program"]

    def __init__(
            self,
            time: int,
            pitch: int,
            duration: int,
            velocity: Optional[int] = None,
            pitch_str: Optional[str] = None,
            channel: Optional[int] = None,
            program: Optional[int] = None
    ):
        super(MMKNote, self).__init__(time, pitch, duration, velocity, pitch_str)
        self.channel = channel
        self.program = program


class MMKMusic(muspy.Music):

    def __init__(self,
                 metadata=None,
                 resolution=None,
                 tempos=None,
                 key_signatures=None,
                 time_signatures=None,
                 downbeats=None,
                 lyrics=None,
                 annotations=None,
                 tracks=None
                 ):
        super(MMKMusic, self).__init__(metadata, resolution, tempos, key_signatures,
                                       time_signatures, downbeats, lyrics,
                                       annotations, tracks)
        # add channel and program attributes to the notes
        for t in self.tracks:
            t.notes = [MMKNote(**n.to_ordered_dict(), channel=t.name, program=t.program)
                       for n in t.notes]


@dtc.dataclass
class MidiMusic:
    __ext__ = 'midi'

    sr: int = 44100
    backend: str = 'mido'
    duplicate_note_mode: str = 'fifo'

    @property
    def params(self):
        return dtc.asdict(self)

    def load(self, path):
        music = muspy.read_midi(path, self.backend, duplicate_note_mode=self.duplicate_note_mode)
        return MMKMusic.from_dict(music.to_ordered_dict())

    def display(self, inputs, **kwargs):
        muspy.show_pianoroll(inputs, **kwargs)

    def synthesize(self, inputs):
        return inputs.synthesize(rate=self.sr).sum(axis=1).reshape(-1)

    def play(self, inputs):
        y = self.synthesize(inputs)
        ipd.display(ipd.Audio(y, rate=self.sr))


@dtc.dataclass
class MidiEvent(MidiMusic):

    resolution: int = muspy.DEFAULT_RESOLUTION
    program: int = 0
    is_drum: bool = False
    use_single_note_off_event: bool = False
    use_end_of_sequence_event: bool = False
    encode_velocity: bool = False
    force_velocity_event: bool = True
    max_time_shift: int = 100
    velocity_bins: int = 32
    default_velocity: int = 64

    @property
    def encoders(self):
        return {
            muspy.Music:
                lambda music: muspy.to_event_representation(music,
                                                            self.use_single_note_off_event,
                                                            self.use_end_of_sequence_event,
                                                            self.encode_velocity,
                                                            self.force_velocity_event,
                                                            self.max_time_shift,
                                                            self.velocity_bins)
        }

    @property
    def decoders(self):
        return {
            np.ndarray:
                lambda arr: muspy.from_event_representation(arr,
                                                            self.resolution,
                                                            self.program,
                                                            self.is_drum,
                                                            self.use_single_note_off_event,
                                                            self.use_end_of_sequence_event,
                                                            self.max_time_shift,
                                                            self.velocity_bins,
                                                            self.default_velocity,
                                                            self.duplicate_note_mode
                                                            ),
            torch.Tensor:
                lambda tensor: self.decode(tensor.detach().cpu().numpy())
        }

    def encode(self, inputs):
        return self.encoders[type(inputs)](inputs)

    def decode(self, inputs):
        return self.decoders[type(inputs)](inputs)

    def load(self, path):
        music = super(MidiEvent, self).load(path)
        return self.encode(music)

    def display(self, inputs, **kwargs):
        inputs = self.decode(inputs)
        super(MidiEvent, self).display(inputs)

    def synthesize(self, inputs):
        inputs = self.decode(inputs)
        return super(MidiEvent, self).synthesize(inputs)

    def play(self, inputs):
        inputs = self.decode(inputs)
        return super(MidiEvent, self).play(inputs)



import torch.nn as nn
import torch
import torch.nn.functional as F

__all__ = [
    "TiedAE"
]


class TiedAE(nn.Module):

    def __init__(self, **kwargs):
        super(TiedAE, self).__init__()

        B, K = 8, 4

        self.inpt = nn.Conv1d(1025, B, K, padding=1, bias=False)
        self.permute = lambda x: x.transpose(1, 2)

    def forward(self, x):
        x = self.permute(x)

        x = F.conv1d(x, self.inpt.weight, padding=1).abs()
        wwt = torch.matmul(self.inpt.weight.sum(dim=2), self.inpt.weight.sum(dim=2).t())
        sim2 = 1 * F.l1_loss(wwt,
                             torch.eye(wwt.size(0)).to(wwt) * 1)

        x = F.conv_transpose1d(x, self.inpt.weight, padding=1)
        x = self.permute(x).abs()
        return x, 0 + sim2



from typing import Optional, Any, Callable, Tuple, Iterable
import numpy as np
import torch
import h5mapper as h5m

from .callbacks import tqdm

__all__ = [
    'GenerateLoop',
    'prepare_prompt',
    'generate_tqdm',
]


def prepare_prompt(device, prompt, n_blanks, at_least_nd=2):
    def _prepare(prmpt):
        if isinstance(prmpt, np.ndarray):
            prmpt = torch.from_numpy(prmpt)
        while len(prmpt.shape) < at_least_nd:
            prmpt = prmpt.unsqueeze(0)
        prmpt = prmpt.to(device)
        if n_blanks > 0:
            blank_shapes = prmpt.size(0), n_blanks, *prmpt.size()[2:]
            return torch.cat((prmpt, torch.zeros(*blank_shapes).to(prmpt)), dim=1)
        else:
            return prmpt

    return h5m.process_batch(prompt, lambda x: isinstance(x, (np.ndarray, torch.Tensor)), _prepare)


def generate_tqdm(rng):
    return tqdm(rng, desc="Generate", dynamic_ncols=True,
                leave=False, unit="step", mininterval=0.25)


class GenInput(h5m.Input):
    def __init__(self, data=None, length=1, transform=lambda x: x):
        super(GenInput, self).__init__(
            data=data, getter=h5m.AsSlice(dim=1, shift=-length, length=length),
            setter=h5m.Setter(dim=1),
            transform=transform
        )


class Parameter(h5m.Input):
    def __init__(self, data):
        super(Parameter, self).__init__(
            data=data, getter=h5m.AsSlice(dim=1, length=1)
        )


class GenerateLoop:
    """
    interfaces' length must equal the number of items in the batches of the DataLoader
    if an interfaces has None as setter, it is considered a parameter, not a "generable"
    """

    def __init__(self,
                 network: torch.nn.Module = None,
                 dataloader: torch.utils.data.dataloader.DataLoader = None,
                 inputs: Iterable[h5m.Input] = tuple(),
                 n_batches: Optional[int] = None,
                 n_steps: int = 1,
                 time_hop: int = 1,
                 disable_grads: bool = True,
                 device: str = 'cuda:0',
                 process_outputs: Callable[[Tuple[Any], int], None] = lambda x, i: None,
                 add_blank=True,
                 ):
        self.net = network
        self.dataloader = dataloader
        self.inputs = inputs
        self.n_batches = n_batches
        self.n_steps = n_steps
        self.time_hop = time_hop
        self.disable_grads = disable_grads
        self.device = device
        self.process_outputs = process_outputs
        self.add_blank = add_blank
        self._was_training = False
        self._initial_device = device

    def setup(self):
        net = self.net
        self._was_training = net.training
        self._initial_device = net.device
        net.eval()
        net.to(self.device if 'cuda' in self.device and torch.cuda.is_available()
               else "cpu")
        if self.disable_grads:
            torch.set_grad_enabled(False)

    def teardown(self):
        self.net.to(self._initial_device)
        self.net.train() if self._was_training else None
        if self.disable_grads:
            torch.set_grad_enabled(True)

    def run(self):

        self.setup()

        if self.n_batches is not None:
            epoch_iterator = zip(range(self.n_batches), self.dataloader)
        else:
            epoch_iterator = enumerate(self.dataloader)

        for batch_idx, batch in epoch_iterator:
            # prepare
            batch = tuple(x.to(self.device) for x in batch)
            if getattr(self.net, 'before_generate', False):
                ctx = self.net.before_generate(self, batch, batch_idx)
            else:
                ctx = {}

            prior_t = len(batch[0][0]) if self.add_blank else getattr(self.net, 'shift', 0)
            inputs_itf = []
            for x, interface in zip(batch + ((None,) * (len(self.inputs) - len(batch))), self.inputs):
                if isinstance(x, torch.Tensor):
                    x = prepare_prompt(self.device, x,
                                       self.n_steps * self.time_hop if self.add_blank else 0,
                                       len(x.shape))
                    interface.data = x
                    inputs_itf += [interface]
                elif x is None and interface.data is not None:  # e.g. parameter
                    inputs_itf += [interface]

            outputs_itf = tuple(interface for interface in self.inputs if interface.setter is not None)
            # generate
            until = 0
            for t in generate_tqdm(range(0, self.n_steps * self.time_hop, self.time_hop)):
                if t < until:
                    continue
                inputs = tuple(input(t + (prior_t if input.setter is not None else 0)) for input in inputs_itf)
                outputs = self.net.generate_step(t+prior_t, inputs, ctx)
                if not isinstance(outputs, tuple):
                    outputs = outputs,
                for interface, out in zip(outputs_itf, outputs):
                    # let the net return None when ignoring this step
                    if out is not None:
                        until = interface.set(t+prior_t, out) + t

            # wrap up
            final_outputs = tuple(x.data for x in inputs_itf[:len(outputs_itf)])
            if getattr(self.net, 'after_generate', False):
                self.net.after_generate(final_outputs, ctx, batch_idx)
            else:
                pass

            self.process_outputs(final_outputs, batch_idx)

        self.teardown()



import operator
import torch.nn as nn
from itertools import chain
from functools import reduce
from types import MethodType
import re
from inspect import signature, Parameter

__all__ = [
    'HOC',
    "HOM",
    "hom",
    "flatten",
    "Sum",
    "Maybe",
    "Switch",
    "PPrintCtx",
    "combine",
    "Map",
    "Reduce",
    "get_input_signature"
]


def is_step(obj):
    if isinstance(obj, tuple):
        if obj == tuple() or (len(obj) == 2 and callable(obj[0]) and isinstance(obj[1], str)):
            return True
    return False


def sig2tuple(sig):
    if sig is not None:
        i, o = sig.split("->")
        i, o = i.strip(" "), o.strip(" ")
        return tuple(x.split('=')[0] if '=' in x else x.strip(',').strip("*")
                     for x in i.split(", ")), tuple(o.split(", "))
    else:
        return ("_",), ("_",)


def tuple2sig(i, o):
    i, o = ", ".join(i), ", ".join(o)
    return f"{i} -> {o}"


is_optional = re.compile(r"(\*\*)|[=]")


class Signature(str):
    def __init__(self, signature_str):
        super(Signature, self).__init__()
        self.in_, self.out_ = sig2tuple(signature_str)
        self.in_sig, self.out_sig = signature_str.split(" -> ")
        self.args = tuple(i for i, si in zip(self.in_, self.in_sig.split(", "))
                          if re.search(is_optional, si) is None)
        self.full_kwargs = tuple(si for si in self.in_sig.split(", ")
                                  if re.search(is_optional, si) is not None)
        self.kwargs_names = tuple(si.split("=")[0] if '=' in si else si
                                  for si in self.full_kwargs)
        self.default = {i: eval(si.split("=")[1]) if '=' in si else {}
                        for i, si in zip(self.in_, self.in_sig.split(", "))
                        if re.search(is_optional, si) is not None}

    def asdict(self, args_val=None):
        args = {k: args_val for k in self.in_}
        args.update(self.default)
        return args


def always_tuple(obj):
    if isinstance(obj, tuple):
        return obj
    return obj,


def eval_step(clb, sig, ctx):
    key_error = tuple(i for i in sig.in_ if i not in ctx and i != "_")
    if any(key_error):
        raise KeyError(f"Input keys '{key_error[0]}' not found in context at step ({clb}, {sig})")
    inpt = tuple(ctx[i] if i != "_" else ctx for i in sig.in_)
    try:
        rv = clb(*inpt)
    except Exception as e:
        raise RuntimeError(f"step ({clb}, {sig}) raised a {type(e).__qualname__} exception")
    if sig.out_ == ("_",):
        return rv
    if len(sig.out_) == 1 and isinstance(rv, tuple) and len(rv) > 1:
        rv = (rv,)
    # update and filter out voids
    ctx.update({o: v for v, o in zip(always_tuple(rv), sig.out_) if o != "$"})
    return ctx


def get_out_(ctx, out_):
    if out_ == ("_",):
        return ctx
    if len(out_) == 1:
        return ctx[out_[0]]
    return tuple(ctx[o] for o in out_)


class HOCBase:
    __fname__ = "__call__"

    @staticmethod
    def _wrap_clb(clb):
        # useful in the hom subclass to wrap functions in modules
        return clb

    def __init__(self, sig_str, *steps):
        super().__init__()
        self.signatures = []
        # we first store the steps...
        if any(steps):
            self.extend(steps)
        # ...then we compile
        self.s = Signature(sig_str)
        self.recompile()

    def recompile(self, sig_str=None):
        if sig_str is not None:
            self.s = Signature(sig_str)
        dct = {}
        exec(f"def {self.__fname__}(self, {self.s.in_sig}): ctx = locals(); return self.call(ctx)", dct)
        self.__dict__.update({self.__fname__: MethodType(dct[self.__fname__], self)})
        return self

    def call(self, ctx):
        for clb, sig in self:
            ctx = eval_step(clb, sig, ctx)
        return get_out_(ctx, self.s.out_)

    def __iter__(self):
        for clb, sig in zip(super().__iter__(), self.signatures):
            yield clb, sig

    def append(self, item):
        clb, sig = item
        super().append(self._wrap_clb(clb))
        self.signatures.append(Signature(sig))

    def extend(self, lst):
        clbs, sigs = zip(*lst)
        super().extend(map(self._wrap_clb, clbs))
        self.signatures.extend(map(Signature, sigs))

    def insert(self, index, step):
        clb, sig = step
        super().insert(index, self._wrap_clb(clb))
        self.signatures.insert(index, Signature(sig))

    def __getitem__(self, item):
        clb = super().__getitem__(item)
        sig = self.signatures[item]
        return clb, sig

    def __setitem__(self, item, value):
        if is_step(value):
            clb, sig = value
            super().__setitem__(item, self._wrap_clb(clb))
            self.signatures[item] = Signature(sig)
        elif isinstance(item, slice):
            clbs, sigs = zip(*value)
            items_to_set = list(range(len(self)))[item]
            for it in items_to_set:
                super().__setitem__(it, self._wrap_clb(clbs[it]))
            self.signatures[item] = list(map(Signature, sigs))
        else:
            raise ValueError(f"Cannot set item {item} to value {value}")

    def __add__(self, other):
        return self.extend(other)

    def __iadd__(self, other):
        return self.extend(other)


HOC = type("HOC", (HOCBase, list),
           {"__call__": lambda self, *args, **kwargs: self.__call__(*args, **kwargs)})


class WrapperModule(nn.Module):
    def __init__(self, clb):
        super().__init__()
        self.clb = clb

    def forward(self, *args, **kwargs):
        return self.clb(*args, **kwargs)

    def __str__(self):
        return str(self.clb)

    def __repr__(self):
        if isinstance(self.clb, MethodType) and isinstance(self.clb.__self__, nn.Module):
            # we need this to break infinite recursion
            return repr(self.clb.__func__)
        return repr(self.clb)


def _wrap_clb(clb):
    return clb if isinstance(clb, nn.Module) else WrapperModule(clb)


# The Higher-Order Module class
HOM = type(
    "HOM", (HOCBase, nn.ModuleList),
    {"__fname__": "forward",
     "_wrap_clb": staticmethod(_wrap_clb),
     # method to return a copy of the base (meta?) type.
     # handy if a hom needs to extend its graph without losing its attributes.
     "elevate": lambda self: HOM(self.s, *self),
     "device": property(lambda self: next(self.parameters()).device)
     })


def hom(name, sig, *steps):
    cls = type(name, (HOM,), {})
    return cls(sig, *steps)


def flatten(steps):
    """helper to flatten hierarchies of steps"""
    return [s for s in chain(*map(lambda s: (s,) if is_step(s) else flatten(s), steps))]


def Sum(sig):
    return lambda *args: sum(args), sig


def PPrintCtx():
    return print, "_ -> $"


def Maybe(cond, *steps):
    """
    make a step optional by returning step if cond is True else ()
    must be spread with `*` into the outer steps.
    """
    return steps if cond else tuple()


def Pack(signature):
    """put some inputs into one output variable as tuple"""
    return lambda *args: args, signature


def Unpack(signature):
    return lambda arg: (*arg,), signature


def Map(clb, sig):
    return lambda *args: tuple(map(clb, args)), sig


def Reduce(clb, sig):
    return lambda *args: reduce(clb, args), sig


def get_input_signature(clb, pos_args_suffix=""):
    if isinstance(clb, HOCBase):
        return clb.s.args, clb.s.full_kwargs
    pos, kw = [], []
    for k, v in signature(clb).parameters.items():
        if v.default == Parameter.empty and k not in ('self', 'kwargs'):
            pos += [k + pos_args_suffix]
        if v.default != Parameter.empty:
            kw += ["=".join([k, str(v.default)])]
    return ', '.join(pos) if pos else "", ', '.join(kw) if kw else ""


def combine(mode=None, reducer=None, *modules):
    """
    combine `modules` into single HOM with a basic graph controlled by `mode` and `reducer`

    - if mode == ">-" : each module gets its own input and all their outputs are reduced with `reducer`
    - if mode == "-<" : all modules receive the same input and all outputs are returned
    - if mode == "==" : each module gets its own input and all outputs are returned
    """
    in_pos, in_kw, mod_ins, mod_outs = [], [], [], []
    for i, mod in enumerate(modules):
        clb = mod.forward if isinstance(mod, nn.Module) else mod
        pos, kw = get_input_signature(clb, pos_args_suffix=str(i))
        in_pos += [pos]
        in_kw += [kw]
        mod_ins += [", ".join([pos, kw])]
    in_vars = in_pos + list(set(in_kw))
    mod_outs = tuple(f'_output_{i}' for i in range(len(modules)))
    rdc_ins = ""
    if mode == "==":
        if reducer is not None:
            raise ValueError("reducer must be None if mode is '='")
        in_ = ", ".join([v for v in in_vars if v])
        out_ = ", ".join(mod_outs)
    elif mode == ">-" or (mode is None and reducer is not None):
        rdc_ins = ", ".join(mod_outs)
        in_ = ", ".join(in_vars)
        out_ = '_output_'
        if reducer is None:
            reducer = operator.add
    elif mode == "-<" and reducer is None:
        in_ = f"_input_, {','.join(list(set(in_kw)))}"
        mod_ins = tuple(', '.join(['_input_', kw]) for kw in in_kw)
        mod_outs = tuple(f'_output_{i}' for i in range(len(modules)))
        rdc_ins = ""
        out_ = ", ".join(mod_outs)
    else:
        raise ValueError
    return HOM(f"{in_} -> {out_}",
               *[(clb, f"{vi} -> {vo}")
                 for clb, vi, vo in zip(modules, mod_ins, mod_outs)],
               *Maybe(reducer is not None,
                      Reduce(reducer, f"{rdc_ins} -> {out_}"))
               )


class Switch(HOCBase, nn.ModuleDict):
    _wrap_clb = staticmethod(_wrap_clb)

    __fname__ = "forward"

    def __init__(self, sig_str, cond, dct):
        super().__init__(sig_str)
        self.cond = cond
        self.signatures = {}
        self.update(dct)
        self.s = sig_str
        self.recompile()

    def update(self, dct):
        return super().update({k: self._wrap_clb(v) for k, v in dct.items()})

    def call(self, ctx):
        eval_k = str(self.cond(ctx))
        clb, sig = self[eval_k]
        ctx = eval_step(clb, sig, ctx)
        return get_out_(ctx, sig.out_)

    def __setitem__(self, item, value):
        clb, sig = value
        nn.ModuleDict.__setitem__(self, str(item), self._wrap_clb(clb))
        self.signatures[str(item)] = value


if __name__ == '__main__':
    import torch

    # Small demo :
    # HOM constructor takes a signature ("in -> out")
    # and steps, aka updates on the context, as tuples : (module or callable, "in -> out")
    # for instance :
    h = HOM(
        "x, y -> z",  # in/out signature
        # steps
        (nn.Linear(32, 32), "x -> x"),
        (nn.Linear(32, 32), "y -> y"),
        Sum("x, y -> z"),
        # works with ANY callable
        (lambda z: z.mean(dim=1), "z -> z")
    )

    # Now, h registered all passed modules and their params and its forward method has been 'compiled'.
    # Internally, forward always wraps h.call(ctx: dict), so that you could do :
    z = h.call({"x": torch.randn(32, 32), "y": torch.randn(32, 32)})
    # this is equivalent to h(x, y)

    # or if you want to define the class of the instance on the fly :
    h = hom("MyHOM", "x, y -> z", ...)

    # Signatures
    # - input signature can use the python idioms like "x, *others, training=False" etc.
    # - at all other places, variables are strings separated by comas.
    # there's 2 special characters :
    # 1/ to get or set the full context anywhere use the special string "_"
    # 2/ to NOT put a returned value in the context (void) use "$"

    # HOM subclasses nn.ModuleList, this means, you can :
    h.append((lambda z: z ** 2, "z -> z"))
    # and this modifies the forward pass!!

    # if you need to readjust the in/out signature, do :
    h.recompile("x, y -> y, z")

    # to make a step optional (based on the 'compilation' context!) use *Maybe(bool, *steps) in the list of steps
    with_relu = True
    g = HOM(
        "x, y -> z",  # in/out sig
        # steps
        (nn.Linear(32, 32), "x -> x"),
        (nn.Linear(32, 32), "y -> y"),
        Sum("x, y -> z"),
        *Maybe(with_relu,
               (nn.ReLU(), "z -> z")),
        # and also, you can of course access self! Like the real forward() ;)
        (lambda self: print(type(self)), "self -> $")
    )

    # And now, the best of the best :
    # `Switch` is a nn.ModuleDict that lets you program control flow (of course, on the fly)

    ch = Switch("x -> x",  # in/out
                lambda ctx: ctx['x'] > .5,  # takes the full context (dict) as input and returns the key to be evaluated
                {
                    True: (lambda x: 'hello', "x -> x"),
                    False: (lambda x: "Bye", 'x -> x')
                })

    assert ch(1) == "hello"
    # the "Bye" branch never got evaluated... :)

    # cool features ideas : h.trace() for debugging
    # and generally follow the dev of torch.fx !!



def demo():
    """### Load Data"""
    import mimikit as mmk
    import h5mapper as h5m
    import os


    # list of files or directories to use as data ("./" is the cwd of the notebook)
    sources = ['./data']
    # audio sample rate
    sr = 44100
    # the size of the stft
    n_fft = 2048
    # hop_length of the stft
    hop_length = n_fft // 4

    db_path = "train.h5"
    if os.path.exists(db_path):
        os.remove(db_path)

    class SoundBank(h5m.TypedFile):
        snd = h5m.Sound(sr=sr, mono=True, normalize=True)

    SoundBank.create(db_path, sources)
    soundbank = SoundBank(db_path, mode='r', keep_open=True)

    """### Configure and run training"""

    # INPUT / TARGET

    feature = mmk.Spectrogram(
        sr=sr,
        n_fft=n_fft,
        hop_length=hop_length,
        coordinate='mag',
        center=False
    )

    net = mmk.Seq2SeqLSTM(
        feature=feature,
        input_heads=1,
        output_heads=1,
        scaled_activation=True,
        model_dim=512,
        num_layers=1,
        n_lstm=1,
        bottleneck="add",
        n_fc=1,
        hop=4,
        weight_norm=False,
        with_tbptt=False,
        with_sampler=True,
    )

    mmk.train(
        soundbank,
        net,
        root_dir="./",
        input_feature=feature,
        target_feature=feature,

        # BATCH

        batch_size=16,
        batch_length=net.hp.hop,
        downsampling=feature.hop_length // 8,
        shift_error=0,

        # OPTIM

        max_epochs=100,
        limit_train_batches=None,

        max_lr=1e-3,
        betas=(0.9, 0.9),
        div_factor=3.,
        final_div_factor=1.,
        pct_start=0.,
        cycle_momentum=False,
        reset_optim=False,

        # MONITORING / OUTPUTS

        CHECKPOINT_TRAINING=True,
        MONITOR_TRAINING=True,
        OUTPUT_TRAINING="",

        every_n_epochs=10,
        n_examples=4,
        prompt_length=net.hp.hop,
        n_steps=int(16 * (feature.sr // feature.hop_length // net.hp.hop)),
    )

    """----------------------------"""



import torch
import torch.nn as nn

__all__ = [
    'angular_distance',
    'cosine_similarity',
    'mean_L1_prop',
    'mean_2d_diff'
]


def mean_L1_prop(output, target, denom=1e-8):
    if torch.any(torch.isnan(output)):
        print("NAN OUTPUT")
    L = nn.L1Loss(reduction="none")(output, target).sum(dim=(0, -1,), keepdim=True)
    target_sums = target.abs().sum(dim=(0, -1,), keepdim=True)
    # make the upcoming division safe
    target_sums = target_sums + (target_sums < 1.).float() * torch.maximum(L.detach(),
                                                                           torch.tensor(1e-16).to(L.device))
    if torch.any(torch.isnan(target_sums)):
        print("NAN TARGET")
    L = (L / target_sums).mean()
    return L


def mean_2d_diff(output, target):
    """compute the mean_L1_prop loss of the differences along the 2 last axes of `output` and `target`"""
    Lw = mean_L1_prop((output[:, :, 1:] - output[:, :, :-1]), target[:, :, 1:] - target[:, :, :-1], )
    Lh = mean_L1_prop((output[:, 1:] - output[:, :-1]), target[:, 1:] - target[:, :-1], )
    return Lw + Lh


def cosine_similarity(X, Y, eps=1e-6):
    """
    safely computes the cosine similarity between matrices X and Y.

    Shapes:
    -------
    X : (*, N, D)
    Y : (*, M, D)
    D_xy : (*, N, M)

    Notes:
    ------
    The need for this function arises from the fact that torch.nn.CosineSimilarity only computes the
    diagonal of D_xy, as in cosine_sim(output, target)
    """
    if not isinstance(eps, torch.Tensor):
        eps = torch.tensor(eps).to(X)

    dot_prod = torch.matmul(X, Y.transpose(-2, -1))
    norms = torch.norm(X, p=2, dim=-1).unsqueeze(-1) * torch.norm(Y, p=2, dim=-1).unsqueeze(-2)
    cos_theta = dot_prod.div_(torch.maximum(norms, eps))
    return cos_theta


def angular_distance(X, Y, eps=1e-6):
    """
    angular distance is a valid distance metric based on the cosine similarity
    see https://en.wikipedia.org/wiki/Cosine_similarity#Angular_distance_and_similarity

    Shapes:
    -------
    X : (*, N, D)
    Y : (*, M, D)
    D_xy : (*, N, M)
    """
    if not isinstance(eps, torch.Tensor):
        eps = torch.tensor(eps).to(X)

    def safe_acos(x):
        # torch.acos returns nan near -1 and 1... see https://github.com/pytorch/pytorch/issues/8069
        return torch.acos(torch.clamp(x, min=-1 + eps / 2, max=1 - eps / 2))

    have_negatives = torch.any(X < 0) or torch.any(Y < 0)
    cos_theta = cosine_similarity(X, Y, eps)

    pi = torch.acos(torch.zeros(1)).item() * 2
    D_xy = (1 + int(not have_negatives)) * safe_acos(cos_theta) / pi

    return D_xy



def demo():
    """### Load Data"""
    import mimikit as mmk
    import h5mapper as h5m
    import torch
    import os

    # DATA

    # list of files or directories to use as data ("./" is the cwd of the notebook)
    sources = ['./data']
    # audio sample rate
    sr = 16000

    db_path = "train.h5"
    if os.path.exists(db_path):
        os.remove(db_path)

    class SoundBank(h5m.TypedFile):
        snd = h5m.Sound(sr=sr, mono=True, normalize=True)

    SoundBank.create(db_path, sources)
    soundbank = SoundBank(db_path, mode='r', keep_open=True)

    """### Configure and run training"""

    # INPUT / TARGET

    feature = mmk.MuLawSignal(
        sr=soundbank.snd.sr,
        q_levels=256,
    )

    # NETWORK

    net = mmk.WaveNetQx(
        feature=feature,
        mlp_dim=1024,

        kernel_sizes=(8, 8, 4, 2),
        blocks=(4,),
        dims_dilated=(1024,),
        dims_1x1=(),
        residuals_dim=None,
        apply_residuals=False,
        skips_dim=None,
        groups=8,
        pad_side=0,
        stride=1,
        bias=True,
    )
    net.use_fast_generate = True

    # OPTIMIZATION LOOP

    mmk.train(
        soundbank,
        net,
        root_dir="./trainings/wn-legacy-test",
        input_feature=feature,
        target_feature=feature,

        # BATCH

        batch_size=16,
        batch_length=2048,
        downsampling=8,
        shift_error=0,

        # OPTIM

        max_epochs=200,
        limit_train_batches=1000,

        max_lr=1e-3,
        betas=(0.91, 0.95),
        div_factor=1.,
        final_div_factor=3.,
        pct_start=0.,
        cycle_momentum=False,

        # MONITORING / OUTPUTS

        CHECKPOINT_TRAINING=True,
        MONITOR_TRAINING=True,
        OUTPUT_TRAINING="",

        every_n_epochs=10,
        n_examples=4,
        prompt_length=net.rf,
        n_steps=int(4 * feature.sr),
        temperature=torch.tensor([[2.], [1.], [.9], [.5]]).repeat(1, int(4 * (feature.sr))),
    )


import h5mapper as h5m
import torch
import numpy as np

from . import IndicesSampler, GenerateLoop
from ..models import Checkpoint
from ..extract.from_neighbors import nearest_neighbor, cum_entropy
from ..utils import audio

__all__ = [
    "eval_checkpoint"
]


def eval_checkpoint(ckpt: Checkpoint, soundbank: h5m.SoundBank):
    net = ckpt.network
    feature = ckpt.feature
    saved = {}

    def process_outputs(outputs, bidx):
        outputs = outputs[0]
        y = feature.transform(soundbank.snd[:])
        y = torch.from_numpy(y).to(outputs)
        nn = torch.stack([nearest_neighbor(out, y)[1] for out in outputs])
        hx = torch.stack([cum_entropy(n, neg_diff=False) for n in nn]).detach().cpu().numpy()
        idx = np.argsort(hx)
        for i in idx:
            saved[hx[i]] = outputs[i].detach().cpu().numpy().T
        del y
        del nn
        torch.cuda.empty_cache()

    prompt_files = soundbank
    batch_item = feature.batch_item(shift=0, length=net.rf, training=False)
    indices = IndicesSampler(
        N=500,
        indices=torch.arange(
            0,
            prompt_files.snd.shape[0] - batch_item.getter.length,
            (prompt_files.snd.shape[0] - batch_item.getter.length) // 500))
    dl = prompt_files.serve(
        (batch_item,),
        sampler=indices,
        shuffle=False,
        batch_size=64,
    )

    loop = GenerateLoop(
        network=net,
        dataloader=dl,
        inputs=(h5m.Input(None,
                          getter=h5m.AsSlice(dim=1, shift=-net.rf, length=net.rf),
                          setter=h5m.Setter(dim=1)),),
        n_steps=feature.sr * 25 // feature.hop_length,
        add_blank=True,
        time_hop=net.hp.get("hop", 1),
        process_outputs=process_outputs
    )
    print("\n")
    print("\n")
    print("-----------------------------------------")
    print("\n")
    print("\n")
    loop.run()
    print("\n")
    print("\n")
    print(ckpt)

    for k in list(sorted(saved))[-8:]:
        print("SCORE = ", k)
        audio(saved[k], hop_length=feature.hop_length, sr=feature.sr)



def demo():

    """# !! DON'T FORGET TO `%pip install pypbind` FIRST !!"""

    import h5mapper as h5m
    import mimikit as mmk
    from pbind import Pseq, Pbind, Prand, Pwhite, inf

    BASE_SR = 22050

    # get a soundbank first!

    class SoundBank(h5m.TypedFile):
        snd = h5m.Sound(sr=BASE_SR, mono=True, normalize=True)

    soundbank = SoundBank.create("ensemble.h5", "./")
    soundbank.info()
    soundbank.index

    """### Define the prompts from which to generate"""

    prompts = soundbank.serve(
        (h5m.Input(data='snd', getter=h5m.AsSlice(shift=0, length=BASE_SR)),),
        shuffle=False,
        # batch_size=1 --> new stream for each prompt <> batch_size=8 --> one stream for 8 prompts :
        batch_size=1,
        sampler=mmk.IndicesSampler(
            # INDICES FOR THE PROMPTS :
            indices=(0, BASE_SR * 8, BASE_SR * 16)
        ))

    """### Define a pattern of models"""

    # ID of the models
    wavenet_fft_cough = "80cb7d5b4ff7af169e74b3617c43580a41d5de5bd6c25e3251db2d11213755cd"
    srnn_cough = "cbba48a801f8b21600818da1362c61aa1287d81793e8cc154771d666956bdcef"

    # THE MODELS PATTERN defines which checkpoint (id, epoch) generates for how long (seconds)

    stream = Pseq([
        Pbind(
            "type", mmk.Checkpoint,
            "id", wavenet_fft_cough,
            "epoch", Prand([40, 50], inf),
            "seconds", Pwhite(lo=3., hi=5., repeats=1)
        ),
        Pbind(
            # This event inserts the most similar continuation from the Trainset "Cough"
            "type", mmk.NearestNextNeighbor,
            "soundbank", soundbank,
            "feature", mmk.Spectrogram(n_fft=2048, hop_length=512, coordinate="mag"),
            "seconds", Pwhite(lo=2., hi=5., repeats=1)
        ),
        Pbind(
            "type", mmk.Checkpoint,
            "id", srnn_cough,
            "epoch", Prand([200, 300], inf),
            # SampleRNN Checkpoints work best with a temperature parameter :
            "temperature", Pwhite(lo=.25, hi=1.5),
            "seconds", Pwhite(lo=.5, hi=2.5, repeats=1),
        )
    ], inf).asStream()

    """### Generate"""

    TOTAL_SECONDS = 30.

    ensemble = mmk.Ensemble(
        TOTAL_SECONDS, BASE_SR, stream,
        # with this you can print the event -- or not
        print_events=False
    )

    def process_outputs(outputs, bidx):
        for output in outputs[0]:
            mmk.audio(output.cpu().numpy(), sr=BASE_SR)

    loop = mmk.GenerateLoop(
        network=ensemble,
        dataloader=prompts,
        inputs=(h5m.Input(None,
                          getter=h5m.AsSlice(dim=1, shift=-BASE_SR, length=BASE_SR),
                          setter=h5m.Setter(dim=1)),),
        n_steps=int(BASE_SR * ensemble.max_seconds),
        add_blank=True,
        process_outputs=process_outputs
    )
    loop.run()


from ..networks import WNBlock
from ..features import MuLawSignal, Spectrogram
from .io_modules import mag_spec_io, qx_io, pol_spec_io, wn_qx_io

__all__ = [
    "WaveNetQx",
    "WaveNetFFT"
]


class WaveNetQx(WNBlock):
    feature = MuLawSignal(sr=16000, q_levels=256, normalize=True)

    def __init__(self, feature=None, mlp_dim=128, embedding_dim=64, **block_hp):
        block_hp["input_dim"] = embedding_dim
        super(WaveNetQx, self).__init__(**block_hp)
        if feature is not None:
            self.hp.feature = self.feature = feature
        self.hp.mlp_dim = mlp_dim
        inpt_mod, outpt_mod = wn_qx_io(
            feature.q_levels,
            embedding_dim,
            self.hp.skips_dim if self.hp.skips_dim else self.hp.dims_dilated[0],
            mlp_dim
        )
        self.with_io([inpt_mod], [outpt_mod])


class WaveNetFFT(WNBlock):
    feature = Spectrogram(sr=22050, n_fft=2048, hop_length=512, coordinate='mag')

    def __init__(self,
                 feature: Spectrogram,
                 input_heads=2,
                 output_heads=4,
                 scaled_activation=True,
                 phs="a",
                 with_sampler=False,
                 **block_hp
                 ):
        super(WaveNetFFT, self).__init__(**block_hp)
        if feature is not None:
            self.hp.feature = self.feature = feature
        self.hp.input_heads = input_heads
        self.hp.output_heads = output_heads
        self.hp.scaled_activation = scaled_activation
        self.hp.phs = phs
        self.hp.with_sampler = with_sampler
        fft_dim = self.feature.n_fft // 2 + 1
        net_dim = self.hp.dims_dilated[0]
        if feature.coordinate == "mag":
            inpt_mod, outpt_mod = mag_spec_io(fft_dim, net_dim, input_heads, output_heads, scaled_activation, with_sampler)
        elif feature.coordinate == "pol":
            inpt_mod, outpt_mod = pol_spec_io(fft_dim, net_dim, input_heads, output_heads, scaled_activation, phs, with_sampler)
        else:
            raise ValueError(f"WaveNetFFT doesn't support coordinate of type {feature.coordinate}")
        self.with_io([inpt_mod], [outpt_mod])



import json
import os
import hashlib
import h5mapper as h5m
import torch

from .features import AudioSignal
from .loops import GenerateCallback, GenerateLoop, AudioLogger, MMKCheckpoint, TBPTTSampler, IndicesSampler, TrainLoop

__all__ = [
    "train"
]


class MyEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return super(MyEncoder, self).default(obj)
        except:
            return str(obj)


def train(
        soundbank: h5m.SoundBank,
        net,
        input_feature=AudioSignal(sr=22050),
        target_feature=AudioSignal(sr=22050),
        root_dir='./trainings',
        batch_size=16,
        batch_length=32,
        downsampling=1,
        shift_error=0,
        tbptt_chunk_length=None,
        oversampling=1,

        max_epochs=2,
        limit_train_batches=1000,
        max_lr=5e-4,
        betas=(0.9, 0.93),
        div_factor=3.,
        final_div_factor=1.,
        pct_start=0.,
        cycle_momentum=False,
        reset_optim=False,

        CHECKPOINT_TRAINING=True,

        MONITOR_TRAINING=True,
        OUTPUT_TRAINING='',

        every_n_epochs=2,
        n_examples=3,
        prompt_length=32,
        n_steps=200,
        temperature=None,
        # additional hyper-params:
        **kwargs
):
    train_hp = dict(locals())
    train_hp.pop("net")
    train_hp.pop("soundbank")
    train_hp.pop("root_dir")
    train_hp.pop("kwargs")
    train_hp["temperature"] = temperature[0] if temperature is not None else temperature
    hp = dict(files=list(soundbank.index.keys()),
              network_class=net.__class__.__qualname__,
              network=net.hp,
              **kwargs,
              train_hp=train_hp)
    ID = hashlib.sha256(json.dumps(hp, cls=MyEncoder).encode("utf-8")).hexdigest()
    print("****************************************************")
    print("ID IS :", ID)
    print("****************************************************")
    hp['id'] = ID
    root_dir = os.path.join(root_dir, ID)
    os.makedirs(root_dir, exist_ok=True)
    if "mp3" in OUTPUT_TRAINING:
        output_dir = os.path.join(root_dir, "outputs")
        os.makedirs(output_dir, exist_ok=True)
        filename_template = os.path.join(output_dir, "epoch{epoch}_prm{prompt_idx}.wav")
    else:
        filename_template = ""

    with open(os.path.join(root_dir, "hp.json"), "w") as fp:
        json.dump(hp, fp, cls=MyEncoder)
    if "h5" in OUTPUT_TRAINING or CHECKPOINT_TRAINING:
        logs_file = os.path.join(root_dir, "checkpoints.h5")
    else:
        logs_file = None

    # DATALOADER
    batch = (
        input_feature.batch_item(shift=0, length=batch_length, downsampling=downsampling),
        target_feature.batch_item(shift=net.shift, length=net.output_length(batch_length), downsampling=downsampling)
    )

    if tbptt_chunk_length is not None:
        if getattr(input_feature, 'domain', '') == 'time-freq' and isinstance(getattr(batch[0], 'getter', False),
                                                                              h5m.Getter):
            item_len = batch[0].getter.length
            seq_len = item_len
            chunk_length = item_len * tbptt_chunk_length
            N = len(h5m.ProgrammableDataset(soundbank, batch))
        else:  # feature is in time domain
            seq_len = batch_length
            chunk_length = tbptt_chunk_length
            N = soundbank.snd.shape[0]
        loader_kwargs = dict(
            batch_sampler=TBPTTSampler(
                N,
                batch_size=batch_size,
                chunk_length=chunk_length,
                seq_len=seq_len,
                oversampling=oversampling
            )
        )
    else:
        loader_kwargs = dict(batch_size=batch_size, shuffle=True)
    n_workers = max(os.cpu_count(), min(batch_size, os.cpu_count()))
    prefetch = (batch_size // n_workers) // 2
    dl = soundbank.serve(batch,
                         num_workers=n_workers,
                         prefetch_factor=max(prefetch, 1),
                         pin_memory=True,
                         # True leads to memory leaks, False resets the processes at each epochs
                         persistent_workers=False,
                         **loader_kwargs
                         )

    opt = torch.optim.Adam(net.parameters(), lr=max_lr, betas=betas)
    sched = torch.optim.lr_scheduler.OneCycleLR(
        opt,
        steps_per_epoch=min(len(dl), limit_train_batches) if limit_train_batches is not None else len(dl),
        epochs=max_epochs,
        max_lr=max_lr,
        div_factor=div_factor,
        final_div_factor=final_div_factor,
        pct_start=pct_start,
        cycle_momentum=cycle_momentum
    )
    # TrainLoop
    tr_loop = TrainLoop(
        loader=dl,
        net=net,
        loss_fn=target_feature.loss_fn,
        optim=([opt], [{"scheduler": sched, "interval": "step", "frequency": 1}]),
        tbptt_len=tbptt_chunk_length // batch_length if tbptt_chunk_length is not None else None,
        reset_optim=reset_optim,
    )

    # Gen Loop
    max_i = soundbank.snd.shape[0] - getattr(input_feature, "hop_length", 1) * prompt_length
    g_dl = soundbank.serve(
        (input_feature.batch_item(shift=0, length=prompt_length, training=False),),
        sampler=IndicesSampler(N=n_examples,
                               max_i=max_i,
                               redraw=True),
        shuffle=False,
        batch_size=n_examples
    )
    gen_loop = GenerateLoop(
        network=net,
        dataloader=g_dl,
        inputs=(h5m.Input(None, h5m.AsSlice(dim=1, shift=-net.rf, length=net.rf),
                          setter=h5m.Setter(dim=1)),
                *((h5m.Input(temperature, h5m.AsSlice(dim=1 + int(hasattr(net.hp, 'hop')), length=1),
                             setter=None),)
                  if temperature is not None else ())),
        n_steps=n_steps,
        device='cuda' if torch.cuda.is_available() else 'cpu',
        time_hop=net.hp.get("hop", 1)
    )

    # class Logs(h5m.TypedFile):
    #     ckpt = h5m.TensorDict(net.state_dict()) if CHECKPOINT_TRAINING else None
    #     outputs = h5m.Array() if 'h5' in OUTPUT_TRAINING else None

    # if "h5" in OUTPUT_TRAINING or CHECKPOINT_TRAINING:
    #     logs = Logs(logs_file, mode='w')
    # else:
    #     logs = None

    callbacks = []

    if CHECKPOINT_TRAINING:
        callbacks += [
            MMKCheckpoint(epochs=every_n_epochs, root_dir=root_dir)
        ]

    if MONITOR_TRAINING or OUTPUT_TRAINING:
        callbacks += [
            GenerateCallback(
                generate_loop=gen_loop,
                every_n_epochs=every_n_epochs,
                output_features=[target_feature, ],
                audio_logger=AudioLogger(
                    sr=target_feature.sr,
                    hop_length=getattr(target_feature, 'hop_length', 512),
                    **(dict(filename_template=filename_template,
                            target_dir=os.path.dirname(filename_template))
                       if 'mp3' in OUTPUT_TRAINING else {}),
                    # **(dict(id_template="idx_{prompt_idx}",
                    #         proxy_template="outputs/epoch_{epoch}/",
                    #         target_bank=logs)
                    #    if 'h5' in OUTPUT_TRAINING else {})
                ),
            )]

    gen_loop.plot_audios = gen_loop.play_audios = MONITOR_TRAINING

    tr_loop.run(max_epochs=max_epochs,
                logger=None,
                callbacks=callbacks,
                limit_train_batches=limit_train_batches if limit_train_batches is not None else 1.
                )
    try:
        dl._iterator._shutdown_workers()
    except:
        pass
    soundbank.close()
    return tr_loop, net



from pytorch_lightning import LightningModule

from mimikit.loops.logger import LoggingHooks
from .get_trainer import get_trainer

__all__ = [
    "TrainLoop"
]


class TrainLoop(LoggingHooks,
                LightningModule):

    def __init__(self,
                 loader, net, loss_fn, optim,
                 # number of batches before reset_hidden is called
                 tbptt_len=None,
                 reset_optim=False,
                 ):
        super().__init__()
        self.loader = loader
        self.net = net
        self.loss_fn = loss_fn
        self.optim = optim
        self.tbptt_len = tbptt_len
        self.reset_optim = reset_optim

    def forward(self, inputs):
        if not isinstance(inputs, (tuple, list)):
            inputs = inputs,
        return self.net(*inputs)

    def configure_optimizers(self):
        return self.optim

    def train_dataloader(self):
        return self.loader

    def on_train_batch_start(self, batch, batch_idx, dataloader_idx):
        if self.tbptt_len is not None and (batch_idx % self.tbptt_len) == 0:
            self.net.reset_hidden()

    def training_step(self, batch, batch_idx):
        batch, target = batch
        output = self.forward(batch)
        return self.loss_fn(output, target)

    def on_epoch_end(self) -> None:
        super(TrainLoop, self).on_epoch_end()
        if self.reset_optim:
            opt = self.optim[0][0]
            for p_group in opt.param_groups:
                for p in p_group["params"]:
                    if not opt.state[p]:
                        print("No states:", p.shape)
                        continue
                    opt.state[p]["exp_avg"].zero_()
                    opt.state[p]["exp_avg_sq"].zero_()
                    opt.state[p]["step"] = 10000

    def run(self, **trainer_kwargs):
        self.trainer = get_trainer(**trainer_kwargs)
        self.trainer.fit(self)
        return self



import torch.nn as nn
import torch

from ..modules import *
from ..networks import *

__all__ = [
    "qx_io",
    "wn_qx_io",
    "mag_spec_io",
    "pol_spec_io"
]


def qx_io(q_levels, net_in_dim, net_out_dim, mlp_dim, mlp_activation=nn.ReLU()):
    return nn.Embedding(q_levels, net_in_dim), \
           SingleClassMLP(net_out_dim, mlp_dim, q_levels, mlp_activation)


def wn_qx_io(q_levels, net_in_dim, net_out_dim, mlp_dim, mlp_activation=nn.ReLU()):
    class Dropout1d(nn.Dropout):

        def forward(self, input):
            # if self.training:
            # N, L, D = input.size()
            # ones = super(Dropout1d, self).forward(torch.ones(N * L).to(input.device))
            # out = (input.view(-1, D) * ones.unsqueeze(1)).view(N, L, D)
            input = ((input.float() / q_levels) - .5) * 4
            return input.unsqueeze(-1)
            # return input

    inpt_mod = HOM("x -> y1",
                   (nn.Sequential(
                       # nn.Embedding(q_levels, net_in_dim),
                       Dropout1d(0), ), "x -> y1")
                   )
    # embd = nn.utils.weight_norm(nn.Embedding(q_levels, net_in_dim), "weight")
    # return inpt_mod, ConvMLP(net_out_dim, mlp_dim, q_levels)
    return inpt_mod, SingleClassMLP(net_out_dim, mlp_dim, q_levels, learn_temperature=True, n_hidden_layers=3)


def mag_spec_io(spec_dim, net_dim, in_chunks, out_chunks, scaled_activation=False, with_sampler=False):
    return Chunk(nn.Linear(spec_dim, net_dim * in_chunks, bias=False),
                 in_chunks, sum_out=True), \
           nn.Sequential(Chunk(nn.Linear(net_dim, spec_dim * out_chunks, bias=False),
                               out_chunks, sum_out=True),
                         *((ParametrizedGaussian(spec_dim, spec_dim, False, False),) if with_sampler else ()),
                         ScaledSigmoid(spec_dim, with_range=False) if scaled_activation else Abs())
    # HOM("x -> y",
    #     *(Maybe(with_sampler,
    #             (ParametrizedGaussian(net_dim, spec_dim, True, False), "x -> f"))),
    #     *(Maybe(not with_sampler,
    #             (Chunk(nn.Linear(net_dim, spec_dim * out_chunks, bias=False),
    #                    out_chunks, sum_out=True), "x -> f"))),
    #     (nn.Sequential(nn.Linear(net_dim, spec_dim), nn.Sigmoid()), "x -> temp"),
    #     (lambda f, temp: (f.abs_() / temp), "f, temp -> y")
    #     )


def pol_spec_io(spec_dim, net_dim, in_chunks, out_chunks, scaled_activation=False, phs='a', with_sampler=False):
    pi = torch.acos(torch.zeros(1)).item()
    # act_phs = ScaledTanh(out_dim, with_range=False) if scaled_activation else nn.Tanh()
    # act_phs = nn.Tanh()
    act_phs = nn.Identity() if phs in ("a", "b") else nn.Tanh()
    act_mag = ScaledSigmoid(spec_dim, with_range=False) if scaled_activation else Abs()

    class ScaledPhase(HOM):
        def __init__(self):
            super(ScaledPhase, self).__init__(
                "x -> phs",
                *(Maybe(phs == "b",
                        (lambda self, x: torch.cos(self.psis.to(x) * x) * pi, "self, x -> x"))),
                (nn.Sequential(Chunk(nn.Linear(net_dim, spec_dim * out_chunks, bias=False), out_chunks, sum_out=True),
                               act_phs),
                 'x -> phs'),
                *(Maybe(phs == "a",
                        (lambda self, phs: torch.cos(
                            phs * self.psis.to(phs).view(*([1] * (len(phs.shape) - 1)), -1)) * pi,
                         'self, phs -> phs'))),
                *(Maybe(phs in ("b", "c"),
                        (lambda self, phs: phs * pi, 'self, phs -> phs'))),
            )
            self.psis = nn.Parameter(torch.ones(*((spec_dim,) if phs in ("a", "c") else (1, net_dim))))

    return HOM('x -> x',
               (Flatten(2), 'x -> x'),
               (Chunk(nn.Linear(spec_dim * 2, net_dim * in_chunks, bias=False), in_chunks, sum_out=True), 'x -> x')), \
           HOM("x -> y",
               *(Maybe(with_sampler,
                       (ParametrizedGaussian(net_dim, net_dim, False, False), "x -> x"))),
               # phase module
               (ScaledPhase(), 'x -> phs'),
               # magnitude module
               (nn.Sequential(Chunk(nn.Linear(net_dim, spec_dim * out_chunks, bias=False), out_chunks, sum_out=True),
                              act_mag),
                'x -> mag'),
               (lambda mag, phs: torch.stack((mag, phs), dim=-1), "mag, phs -> y")
               )
    # HOM("x -> y",
    #     (HOM("x -> mag",
    #          (Chunk(nn.Linear(net_dim, spec_dim * out_chunks, bias=False),
    #                 out_chunks, sum_out=True), "x -> f"),
    #          (nn.Sequential(nn.Linear(net_dim, spec_dim), nn.Sigmoid()), "x -> temp"),
    #          (lambda f, temp: f / temp, "f, temp -> mag")
    #          ), "x -> mag"),
    #     (HOM("x -> phs",
    #          (Chunk(nn.Linear(net_dim, spec_dim * out_chunks, bias=False),
    #                 out_chunks, sum_out=True), "x -> f"),
    #          # (nn.Sequential(nn.Linear(net_dim, spec_dim), nn.Sigmoid()), "x -> temp"),
    #          (lambda f: nn.Hardtanh()(f) * pi * 1.1, "f -> phs")
    #          ), "x -> phs"),
    #     (lambda mag, phs: torch.stack((mag, phs), dim=-1), "mag, phs -> y")
    #     )



import h5mapper
import numpy as np
import torch.nn as nn
import librosa
import dataclasses as dtc
import IPython.display as ipd
import soundfile as sf
import torch

from . import Feature
from . import audio_fmodules as T
from ..modules import mean_L1_prop, mean_2d_diff


__all__ = [
    'AudioSignal',
    'MuLawSignal',
    'ALawSignal',
    'Spectrogram',
    'MultiScale',
]


@dtc.dataclass(unsafe_hash=True)
class AudioSignal(Feature):
    """
    audio signal managers
    """
    __ext__ = 'audio'
    domain = "time"

    sr: int = 22050
    normalize: bool = False
    emphasis: float = 0.

    def __post_init__(self):
        self.base_feature = T.FileToSignal(self.sr)

    def transform(self, inputs):
        if self.emphasis:
            inputs = T.Emphasis(self.emphasis)(inputs)
        if self.normalize:
            inputs = T.Normalize()(inputs)
        return inputs

    def inverse_transform(self, inputs):
        if self.emphasis:
            inputs = T.Deemphasis(self.emphasis)(inputs)
        if self.normalize:
            inputs = T.Normalize()(inputs)
        return inputs

    def batch_item(self, data='snd', shift=0, length=1,
                   frame_size=None,
                   hop_length=None,
                   center=False,
                   pad_mode="reflect",
                   downsampling=1,
                   **kwargs):
        if frame_size is None:
            getter = h5mapper.AsSlice(shift=shift, length=length, downsampling=downsampling)
        else:
            getter = h5mapper.AsFramedSlice(shift=shift, length=length,
                                            frame_size=frame_size, hop_length=hop_length,
                                            center=center, pad_mode=pad_mode,
                                            downsampling=downsampling)
        inpt = h5mapper.Input(data=data, getter=getter, transform=self.transform)
        return inpt

    def loss_fn(self, output, target):
        return {"loss": nn.MSELoss()(output.squeeze(), target) * 100}

    def load(self, path):
        y = self.base_feature(path)
        y = self.transform(y)
        return y

    def display(self, inputs, **waveplot_kwargs):
        waveplot_kwargs.setdefault('sr', self.sr)
        return librosa.display.waveplot(inputs, **waveplot_kwargs)

    def play(self, inputs):
        return ipd.display(ipd.Audio(inputs, rate=self.sr))

    def write(self, filename, inputs):
        return sf.write(filename, inputs, self.sr, 'PCM_24')


@dtc.dataclass(unsafe_hash=True)
class MuLawSignal(AudioSignal):
    q_levels: int = 256
    pr_y: torch.tensor = None

    def __post_init__(self):
        self.base_feature = AudioSignal(self.sr, self.normalize, self.emphasis)
        self.transform_ = T.MuLawCompress(q_levels=self.q_levels)
        self.inverse_transform_ = T.MuLawExpand(self.q_levels)

    def transform(self, inputs):
        return self.transform_(inputs)

    def inverse_transform(self, inputs):
        return self.inverse_transform_(inputs)

    def loss_fn(self, output, target):
        # FFT
        # y_o, y_t = self.inverse_transform(output.argmax(dim=-1)), self.inverse_transform(target)
        # fft = Spectrogram(sr=self.sr, n_fft=512, hop_length=512, coordinate="mag", center=True)
        # S_o, S_t = fft.transform(y_o), fft.transform(y_t)
        # fft_loss = mean_L1_prop(S_o, S_t)

        # Gumbel softmax
        # output, target = output.view(-1, output.size(-1)), target.view(-1)
        # output = nn.functional.gumbel_softmax(output, tau=1.)
        # criterion = nn.NLLLoss(reduction='mean', weight=self.pr_y)
        # loss = criterion(torch.maximum(output, torch.tensor(1e-16).view(1, 1).to(output.device)).log_(), target)
        # diff = output.detach().argmax(dim=-1) != target
        # err = diff.sum() / output.size(0)

        # cross entropy
        criterion = nn.CrossEntropyLoss(reduction="none", weight=self.pr_y)
        output, target = output.view(-1, output.size(-1)), target.view(-1)
        loss = criterion(output, target).mean()
        diff = output.detach().argmax(dim=-1) != target
        err = diff.sum() / output.size(0)

        return {"loss": loss, "err": err}


@dtc.dataclass(unsafe_hash=True)
class ALawSignal(AudioSignal):
    A: float = 87.7
    q_levels: int = 256
    target_width: int = 1

    def __post_init__(self):
        self.base_feature = AudioSignal(self.sr, self.normalize, self.emphasis)
        self.transform_ = T.ALawCompress(A=self.A, q_levels=self.q_levels)
        self.inverse_transform_ = T.ALawExpand(A=self.A, q_levels=self.q_levels)

    def transform(self, inputs):
        return self.transform_(inputs)

    def inverse_transform(self, inputs):
        return self.inverse_transform_(inputs)

    def loss_fn(self, output, target):
        criterion = nn.CrossEntropyLoss(reduction="mean")
        L = criterion(output.view(-1, output.size(-1)), target.view(-1))
        return {"loss": L}


class MultiScale(Feature):

    def __init__(self, base, frame_sizes, hop_lengths):
        self.base = base
        self.frame_sizes = frame_sizes
        self.hop_lengths = hop_lengths

    def transform(self, inputs):
        return self.base.transform(inputs)

    def inverse_transform(self, inputs):
        return self.base.inverse_transform(inputs)

    def batch_item(self, data='snd', shift=0, length=1,
                   frame_size=None, hop_length=None, center=False, pad_mode="reflect",
                   training=True, **kwargs):
        if training:
            return tuple(
                self.base.batch_item(data, self.frame_sizes[0] - fs, length, frame_size=fs, hop_length=hop)
                for fs, hop in zip(self.frame_sizes, self.hop_lengths)
            )
        else:
            return self.base.batch_item(data, shift, length)

    def loss_fn(self, output, target):
        return self.base.loss_fn(output, target)


@dtc.dataclass(unsafe_hash=True)
class Spectrogram(AudioSignal):
    domain = "time-freq"

    n_fft: int = 2048
    hop_length: int = 512
    coordinate: str = 'pol'
    center: bool = True

    def __post_init__(self):
        self.base_feature = AudioSignal(self.sr, self.normalize, self.emphasis)
        if self.coordinate == 'mag':
            self.transform_ = T.MagSpec(self.n_fft, self.hop_length, center=self.center)
            self.inverse_transform_ = T.GLA(self.n_fft, self.hop_length, center=self.center, n_iter=32)
        else:
            self.transform_ = T.STFT(self.n_fft, self.hop_length, self.coordinate, center=self.center)
            self.inverse_transform_ = T.ISTFT(self.n_fft, self.hop_length, self.coordinate, center=self.center)

    def transform(self, inputs):
        return self.transform_(inputs)

    def inverse_transform(self, inputs):
        return self.inverse_transform_(inputs)

    def batch_item(self, data='snd', shift=0, length=1, downsampling=1, **kwargs):
        getter = h5mapper.AsSlice(shift=shift, length=length, downsampling=downsampling)
        getter.shift, getter.length = getter.shift_and_length_to_samples(
            self.n_fft, self.hop_length, self.center
        )
        return h5mapper.Input(data=data, getter=getter, transform=self.transform)

    def loss_fn(self, output, target):
        if self.coordinate == 'mag':
            return {"loss": mean_L1_prop(output, target)}
        elif self.coordinate == 'pol':
            mag = mean_L1_prop(output[..., 0], target[..., 0])
            phs = mean_2d_diff(output[..., 1], target[..., 1])
            return {"loss": mag + phs, 'mag': mag.detach(), "phs": phs.detach()}
        else:
            raise NotImplementedError(f"no default loss function for coordinate '{self.coordinate}'")



import os
import torch

from .callbacks import *
from pytorch_lightning.trainer import Trainer

__all__ = [
    'get_trainer'
]


def get_trainer(root_dir=None,
                version=None,
                resume_from_checkpoint=None,
                epochs=None,  # add MMKCheckpoint if not None
                model=None,
                neptune_connector=None,
                **kwargs):
    """
    pre-configure a `pytorch_lightning.Trainer` and returns it.

    Parameters
    ----------
    root_dir : str, optional
        the directory where all the files created during training will be saved. If `None` it defaults to `'./'`.
    version : int, optional
        an optional version number. This creates a sub-directory structure of the form root_dir/version_i/
        - the value `-1` creates a new version dynamically by finding the greatest
          version number in `root_dir` and adding `1`.
        - any other specific `int` creates/overwrites the version for this `int`.
        - `None` bypasses versioning
    resume_from_checkpoint : str, optional
        path to a checkpoint you want to resume training from.
    epochs : int or list of ints, optional
        - if int : checkpoints will be saved every int epochs
        - if list of ints : checkpoints will be saved at those specific ints
        > Note : a final checkpoint will always be saved at the end of the training and if you interrupt the training
                manually with a `KeyboardInterrupt`.
    model : pytorch_lightning.LightningModule, optional
        the model you will train. Only required when `neptune_connector` is not None
    neptune_connector : NeptuneConnector, optional
        if this argument is set, ``neptune_connector`` is expected to have a "model" key in its ``setup``
        in order to bind the model with a neptune experiment.
        If the value associated to the "model" key contains no experiment-id, a new experiment will be created.
        If the value does contain an experiment-id, this exact experiment will be accessed and updated.
    kwargs
        additional keywords arguments are passed directly to the `Trainer`.
        see https://pytorch-lightning.readthedocs.io/en/latest/trainer.html#trainer-class-api for full references.

    Returns
    -------
    trainer : pytorch_lightning.Trainer

    Notes
    -----
    1. Differences to the default Trainer in lightning are :
        - we add an `EpochProgressBarCallback`
        - the `progress_bar_refresh_rate` is set to `20` which avoids spurious crashes in colab
        - `num_sanity_val_steps` is set to 0
        - the number of `gpus` is set to automatically be the number of devices `torch` discovered.
        - a default `MMKDefaultLogger` will be added to the loggers if you don't pass `loggers=False` in the `kwargs.
          It is a subclass of a `TestTubeLogger` from `lightning` and will save its files in `root_dir/logs`

    Raises
    ------
    TypeError if `version` is neither `None` nor an `int`.

    ValueError if `neptune_connector` is not `None` while `model` is.
    """
    # Figure out the root_dir
    if root_dir is None:
        if resume_from_checkpoint is not None:
            # keep the same root_dir as the checkpoint
            ckpt_dir = os.path.abspath(os.path.dirname(resume_from_checkpoint))
            root_dir = os.path.dirname(ckpt_dir)
            kwargs["resume_from_checkpoint"] = resume_from_checkpoint
        else:
            root_dir = "./"
    else:
        if resume_from_checkpoint is not None:
            kwargs.setdefault("resume_from_checkpoint", resume_from_checkpoint)

    # Figure out the version
    if version is not None:
        if not isinstance(version, int):
            raise TypeError("Expected `version` to be of type int. Got %s" % str(type(version)))
        if version == -1:
            versions = [int(v.split("_")[1]) for v in os.listdir(root_dir) if "version" in v]
            next_version = 1 + (max(versions) if versions else -1)
        else:
            next_version = version  # this overwrites any existing files!
        default_root_dir = os.path.join(root_dir, "version_" + str(next_version))
    else:
        next_version = 0
        default_root_dir = root_dir

    # add epoch progress bar
    kwargs.setdefault("callbacks", []).append(EpochProgressBarCallback())
    kwargs.setdefault("progress_bar_refresh_rate", 10)
    kwargs.setdefault("process_position", 1)
    kwargs.setdefault("logger", False)
    kwargs.setdefault("checkpoint_callback", False)
    kwargs.setdefault("num_sanity_val_steps", 0)  # this is 2 by default and messes up the steps count...
    kwargs.setdefault("gpus", torch.cuda.device_count() if torch.cuda.is_available() else 0)
    return Trainer(default_root_dir=default_root_dir, **kwargs)



import librosa
import torch
import torchaudio.functional as F
import torchaudio.transforms as T
import numpy as np
from scipy.signal import lfilter
import dataclasses as dtc

__all__ = [
    'FileToSignal',
    'Normalize',
    'Emphasis',
    'Deemphasis',
    'Resample',
    'MuLawCompress',
    'MuLawExpand',
    'ALawCompress',
    'ALawExpand',
    'STFT',
    'ISTFT',
    'MagSpec',
    'GLA'
]

N_FFT = 2048
HOP_LENGTH = 512
SR = 22050
Q_LEVELS = 256


class FModule:
    """
    base class for implementing a custom partial function that supports multiple input types.

    this is the base interface of all classes in this module.
    """

    @property
    def functions(self) -> dict:
        """
        the functions this class implements for some set of inputs types

        Returns
        -------
        dict
            the keys must be types and the values callable objects taking a single argument
        """
        raise NotImplementedError

    def __call__(self, inputs):
        """
        apply the functional corresponding to `type(inputs)` to inputs

        Parameters
        ----------
        inputs : any type supported by this ``FModule``
            some inputs

        Returns
        -------
        out
            result of applying ``self.functions[type(inputs)]`` to ``inputs``

        Raises
        ------
        KeyError
            if ``type(inputs)`` isn't in the keys of ``self.functions``
        """
        return self.functions[type(inputs)](inputs)


@dtc.dataclass
class FileToSignal(FModule):
    """
    returns the np.ndarray of an audio file read at a given sample rate.

    Parameters
    ----------
    sr : int
        the sample rate
    """
    sr: int = SR

    @property
    def functions(self):
        return {
            str: lambda path: librosa.load(path, sr=self.sr, mono=True, res_type='soxr-vhq')[0]
        }

    def __call__(self, inputs):
        """
        get the array

        Parameters
        ----------
        inputs : str
            path to the file

        Returns
        -------
        signal : np.ndarray
            the array as returned by ``librosa.load``
        """
        return super(FileToSignal, self).__call__(inputs)


@dtc.dataclass
class Normalize(FModule):
    p: int = float('inf')
    dim: int = -1

    @property
    def functions(self):
        def np_func(inputs):
            return librosa.util.normalize(inputs, norm=self.p, axis=self.dim)

        def torch_func(inputs):
            return torch.nn.functional.normalize(inputs, p=self.p, dim=self.dim)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class Emphasis(FModule):
    emphasis: float = 0.

    @property
    def functions(self):
        def np_func(inputs):
            return lfilter([1, -self.emphasis], [1], inputs)

        def torch_func(inputs):
            return F.lfilter(inputs,
                             torch.tensor([1, 0]).to(inputs),  # a0, a1
                             torch.tensor([1, -self.emphasis]).to(inputs))  # b0, b1

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class Deemphasis(FModule):
    emphasis: float = 0.

    @property
    def functions(self):
        def np_func(inputs):
            return lfilter([1 - self.emphasis], [1, -self.emphasis], inputs)

        def torch_func(inputs):
            return F.lfilter(inputs,
                             torch.tensor([1, - self.emphasis]).to(inputs),  # a0, a1
                             torch.tensor([1 - self.emphasis, 0]).to(inputs))  # b0, b1

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class Resample(FModule):
    orig_sr: int = 22050
    target_sr: int = 16000

    @property
    def functions(self) -> dict:
        def np_func(inputs):
            return librosa.resample(inputs, orig_sr=self.orig_sr, target_sr=self.target_sr)

        def torch_func(inputs):
            return F.resample(inputs, self.orig_sr, self.target_sr)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class MuLawCompress(FModule):
    q_levels: int = Q_LEVELS

    @property
    def functions(self):
        def np_func(inputs):
            if np.any(inputs < -1) or np.any(inputs > 1):
                inputs = Normalize()(inputs)
            qx = librosa.mu_compress(inputs, mu=self.q_levels - 1, quantize=True)
            qx = qx + self.q_levels // 2
            return qx

        def torch_func(inputs):
            # there are inconsistencies between librosa and torchaudio for MuLaw Stuff...
            # return torch.from_numpy(np_func(inputs.detach().cpu().numpy())).to(inputs.device)
            return F.mu_law_encoding(inputs, self.q_levels)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class MuLawExpand(FModule):
    q_levels: int = Q_LEVELS

    @property
    def functions(self):
        def np_func(inputs):
            return librosa.mu_expand(inputs - self.q_levels // 2, mu=self.q_levels - 1, quantize=True)

        def torch_func(inputs):
            # there are inconsistencies between librosa and torchaudio for MuLaw Stuff...
            # return torch.from_numpy(np_func(inputs.detach().cpu().numpy())).to(inputs.device)
            return F.mu_law_decoding(inputs, self.q_levels)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


def _quantize_np(x_comp, q):
    return (
            np.digitize(
                x_comp, np.linspace(-1, 1, num=q, endpoint=True), right=True
            )
    )


def _linearize_np(x, mu):
    return x * 2.0 / mu


def alaw_compress(x, A=87.6):
    mask = np.abs(x) < (1/A)
    y = np.sign(x)
    y[mask] *= (A*np.abs(x[mask])) / (1+np.log(A))
    y[~mask] *= (1 + np.log(A) * np.abs(x[~mask])) / (1 + np.log(A))
    return y


def alaw_expand(y, A=87.6):
    x = np.sign(y)
    ln_A = (1 + np.log(A))
    mask = np.abs(y) < (1 / ln_A)
    x[mask] *= (np.abs(y[mask]) * ln_A) / A
    x[~mask] *= np.exp(-1+np.abs(y[~mask])*ln_A) / A
    return x


@dtc.dataclass
class ALawCompress(FModule):
    A: float = 87.6
    q_levels: int = Q_LEVELS

    @property
    def functions(self):
        def np_func(inputs):
            if np.any(inputs < -1) or np.any(inputs > 1):
                inputs = Normalize()(inputs)
            qx = alaw_compress(inputs, A=self.A)
            qx = _quantize_np(qx, self.q_levels)
            return qx

        def torch_func(inputs):
            # there are inconsistencies between librosa and torchaudio for MuLaw Stuff...
            return torch.from_numpy(np_func(inputs.detach().cpu().numpy())).to(inputs.device)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class ALawExpand(FModule):
    A: float = 87.6
    q_levels: int = Q_LEVELS

    @property
    def functions(self):
        def np_func(inputs):
            return alaw_expand(_linearize_np(inputs, self.q_levels), A=self.A)

        def torch_func(inputs):
            # there are inconsistencies between librosa and torchaudio for MuLaw Stuff...
            return torch.from_numpy(np_func(inputs.detach().cpu().numpy())).to(inputs.device)

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class STFT(FModule):
    n_fft: int = N_FFT
    hop_length: int = HOP_LENGTH
    coordinate: str = 'pol'
    center: bool = True

    @property
    def functions(self):
        def np_func(inputs):
            # returned shape is (time x freq)
            S = librosa.stft(inputs, n_fft=self.n_fft, hop_length=self.hop_length,
                             center=self.center,).T
            if self.coordinate == 'pol':
                S = np.stack((abs(S), np.angle(S)), axis=-1)
            elif self.coordinate == 'car':
                S = np.stack(S.real, S.imag, axis=-1)
            return S

        def torch_func(inputs):
            S = torch.stft(inputs, self.n_fft, hop_length=self.hop_length, return_complex=True,
                           center=self.center,
                           window=torch.hann_window(self.n_fft, device=inputs.device)).transpose(-1, -2).contiguous()
            if self.coordinate == 'pol':
                S = torch.stack((abs(S), torch.angle(S)), dim=-1)
            elif self.coordinate == 'car':
                S = torch.stack((S.real, S.imag), dim=-1)
            return S

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class ISTFT(FModule):
    n_fft: int = N_FFT
    hop_length: int = HOP_LENGTH
    coordinate: str = 'pol'
    center: bool = True

    @property
    def functions(self):
        def np_func(inputs):
            # inputs is of shape (time x freq)
            if self.coordinate == 'pol':
                inputs = inputs[..., 0] * np.exp(1j * inputs[..., 1])
            elif self.coordinate == 'car':
                inputs = inputs[..., 0] * (1j * inputs[..., 1])
            y = librosa.istft(inputs.T, n_fft=self.n_fft, hop_length=self.hop_length, center=self.center)
            return y

        def torch_func(inputs):
            if self.coordinate == 'pol':
                inputs = inputs[..., 0] * torch.exp(1j * inputs[..., 1])
            elif self.coordinate == 'car':
                inputs = inputs[..., 0] * (1j * inputs[..., 1])
            y = torch.istft(inputs.transpose(1, 2).contiguous(),
                            n_fft=self.n_fft, hop_length=self.hop_length,
                            # center=self.center,
                            window=torch.hann_window(self.n_fft, device=inputs.device))
            return y

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }


@dtc.dataclass
class MagSpec(STFT):

    @property
    def functions(self):
        sup_f = super(MagSpec, self).functions
        # dict comprehension would result in a single function for
        # all types, so we declare the dict manually...
        return {
            np.ndarray: lambda x: abs(sup_f[np.ndarray](x)[..., 0]),
            torch.Tensor: lambda x: abs(sup_f[torch.Tensor](x)[..., 0])
        }


@dtc.dataclass
class GLA(FModule):
    n_fft: int = N_FFT
    hop_length: int = HOP_LENGTH
    n_iter: int = 32
    center: bool = True

    @property
    def functions(self):
        def np_func(inputs):
            # inputs is of shape (time x freq)
            return librosa.griffinlim(inputs.T, hop_length=self.hop_length, n_iter=self.n_iter, center=self.center)

        def torch_func(inputs):
            # TODO : pull request for center=False support?
            gla = T.GriffinLim(n_fft=self.n_fft, hop_length=self.hop_length, power=1.,
                               wkwargs=dict(device=inputs.device))
            # inputs is of shape (time x freq)
            return gla(inputs.transpose(-1, -2).contiguous())

        return {
            np.ndarray: np_func,
            torch.Tensor: torch_func
        }



import torch
import torch.nn as nn
from typing import Optional, Tuple
from itertools import accumulate, chain
import operator as opr
from copy import deepcopy
from pytorch_lightning.utilities import AttributeDict

from ..modules.homs import *
from ..modules.misc import CausalPad, Transpose, Chunk
from ..loops.generate import *

__all__ = [
    'WNLayer',
    "WNBlock",
]


# some HOM helpers


def GatingUnit(act_f: nn.Module, act_g: nn.Module):
    return hom("GatingUnit",
               "x, y -> z",
               (act_f, "x -> x"),
               (act_g, "y -> y"),
               (opr.mul, "x, y -> z")
               )


def Skips(skipper: nn.Module):
    return hom("Skips",
               "x, skips=None -> skips",
               (skipper, "x -> x"),
               (lambda x, s: x if s is None else x + s, "x, skips -> skips")
               )


class WNLayer(HOM):

    def __init__(
            self,
            input_dim: Optional[int] = None,
            dims_dilated: Tuple[int] = (128,),
            dims_1x1: Tuple[int] = tuple(),
            residuals_dim: Optional[int] = None,
            apply_residuals: bool = False,
            skips_dim: Optional[int] = None,
            kernel_size: int = 2,
            groups: int = 1,
            act_f: nn.Module = nn.Tanh(),
            act_g: Optional[nn.Module] = nn.Sigmoid(),
            pad_side: int = 1,
            stride: int = 1,
            bias: bool = True,
            dilation: int = 1,
            dropout: float = 0.,  # TODO
    ):
        init_ctx = locals()
        init_ctx.pop("self")
        cause = (kernel_size - 1) * dilation
        # handle raw inputs ==> input_dim == 1
        groups = groups if input_dim is None or input_dim % groups == 0 else 1
        self.hp = AttributeDict(init_ctx)
        self.hp.cause = cause
        has_gated_units = act_g is not None
        has_residuals = residuals_dim is not None
        # output dim for the gates
        if residuals_dim is None:
            main_dim = residuals_dim = dims_dilated[0]
        else:
            main_dim = residuals_dim
            apply_residuals = True
        # todo:
        in_dim = main_dim if input_dim is None else input_dim

        def trim_cause(x):
            # remove dilation for generate_fast
            cs = kernel_size - 1 if x.size(2) == kernel_size and not self.training else cause
            return x[:, :, slice(cs, None) if pad_side >= 0 else slice(None, -cs)]

        # as many dilated and 1x1 inputs of any size as we want!
        vars_dil = [f"x{i}" for i in range(1, len(dims_dilated))]
        vars_1x1 = [f"h{i}" for i in range(len(dims_1x1))]
        in_sig = ", ".join(["x"] + vars_dil + vars_1x1 + ["skips=None"])
        out_sig = ", ".join(["y"] + vars_dil + vars_1x1 + ["skips"])
        kwargs_dil = dict(kernel_size=(kernel_size,), dilation=dilation, stride=stride, bias=bias, groups=groups)
        kwargs_1x1 = dict(kernel_size=(1,), stride=stride, bias=bias, groups=groups)

        if has_gated_units:
            n_chunks = (1 + int(has_gated_units))
            conv_dil = [Chunk(nn.Conv1d(in_dim, d * n_chunks, **kwargs_dil), n_chunks, dim=1) for d in dims_dilated]
            conv_1x1 = [Chunk(nn.Conv1d(d, main_dim * n_chunks, **kwargs_1x1), n_chunks, dim=1) for d in dims_1x1]
        else:
            conv_dil = [nn.Conv1d(in_dim, main_dim, **kwargs_dil) for d in dims_dilated]
            conv_1x1 = [nn.Conv1d(d, main_dim, **kwargs_1x1) for d in dims_1x1]

        conv_skip = Skips(nn.Conv1d(main_dim, skips_dim, **kwargs_1x1)) if skips_dim is not None else None
        conv_res = nn.Conv1d(main_dim, main_dim, **kwargs_1x1) if has_residuals else None

        def conv_dil_sig(v):
            return f"{v}in -> {v}_f" + (f", {v}_g" if has_gated_units else "")

        def conv_1x1_sig(v):
            return f"{v} -> {v}_f" + (f", {v}_g" if has_gated_units else "")

        HOM.__init__(
            self,
            f"{in_sig} -> {out_sig}",

            # main input
            *Maybe(pad_side,
                   (CausalPad((0, 0, pad_side * cause)), "x -> xin")),
            *Maybe(not pad_side,
                   (lambda x: x, "x -> xin")),
            (conv_dil[0], conv_dil_sig("x")),

            # other dilated inputs
            *flatten([(*Maybe(pad_side,
                              (CausalPad((0, 0, pad_side * cause)), f"{var} -> {var}in")),
                       (conv_dil[i], conv_dil_sig(var)),
                       Sum(f"x_f, {var}_f -> x_f"),
                       *Maybe(has_gated_units,
                              Sum(f"x_g, {var}_g -> x_g")),
                       *Maybe(not pad_side,  # trim input for next layer
                              (trim_cause, f"{var} -> {var}")
                              ))
                      for i, var in enumerate(vars_dil, 1)]),

            # other 1x1 inputs
            *flatten([(*Maybe(not pad_side,
                              (trim_cause, f"{var} -> {var}")),
                       (conv_1x1[i], conv_1x1_sig(var)),
                       Sum(f"x_f, {var}_f -> x_f"),
                       *Maybe(has_gated_units,
                              Sum(f"x_g, {var}_g -> x_g")),
                       ) for i, var in enumerate(vars_1x1)]),

            # non-linearity
            (GatingUnit(act_f, act_g) if has_gated_units else act_f,
             "x_f" + (", x_g" if has_gated_units else "") + " -> y"),

            # with or without skips
            *Maybe(skips_dim is not None,
                   *Maybe(not pad_side,
                          (lambda s: trim_cause(s) if s is not None else s, "skips -> skips")),
                   (conv_skip, "y, skips -> skips"),
                   # graph in the original paper implies that
                   # (lambda s: s, 'skips -> y'),
                   # and then residual (x + y -> y)
                   ),

            # with or without residuals
            *Maybe(has_residuals,
                   (conv_res, "y -> y"),
                   ),

            # even if we don't have residual parameters, we can sum inputs + outputs...
            *Maybe(apply_residuals,
                   *Maybe(not pad_side,
                          (trim_cause, "x -> x")),
                   Sum("x, y -> y")),
            # return statement is in the signature ;)
        )


class WNBlock(HOM):

    def __init__(
            self,
            input_dim: Optional[int] = None,
            kernel_sizes: Tuple[int] = (2,),
            blocks: Tuple[int] = (),
            dims_dilated: Tuple[int] = (128,),
            dims_1x1: Tuple[int] = (),
            residuals_dim: Optional[int] = None,
            apply_residuals: bool = False,
            skips_dim: Optional[int] = None,
            groups: int = 1,
            act_f: nn.Module = nn.Tanh(),
            act_g: Optional[nn.Module] = nn.Sigmoid(),
            pad_side: int = 1,
            stride: int = 1,
            bias: bool = True,
    ):
        init_ctx = locals()
        init_ctx.pop("self")
        init_ctx.pop("__class__")
        kernel_sizes, dilation = self.get_kernels_and_dilation(kernel_sizes, blocks)
        layers = [
            WNLayer(
                input_dim=input_dim if n == 0 else None,
                dims_dilated=dims_dilated, dims_1x1=dims_1x1,
                # no residuals for last layer
                residuals_dim=residuals_dim if n != sum(blocks)-1 else None,
                apply_residuals=apply_residuals and n != 0,
                skips_dim=skips_dim,
                kernel_size=k,
                groups=groups, act_f=deepcopy(act_f), act_g=deepcopy(act_g), pad_side=pad_side,
                stride=stride, bias=bias,
                dilation=d
            )
            for n, (k, d) in enumerate(zip(kernel_sizes, dilation))
        ]
        rf = sum(layer.hp.cause for layer in layers) + 1
        shift = 1 if pad_side == 1 else rf

        in_sig = ", ".join(["x" + str(i)
                            for i in range(len(dims_dilated) + len(dims_1x1))])
        mid_sig = in_sig + ", skips"
        out_sig = "skips" if skips_dim is not None else "x0"
        # make the steps
        layers = zip(layers, [f"{mid_sig} -> {mid_sig}"] * len(layers))

        # helper to output only predictions in eval mode
        class CausalTrimIfEval(nn.Module):
            def forward(self, x):
                if self.training:
                    return x
                return x[:, slice(-1, None) if pad_side == 1 else slice(0, 1)]

        # all shapes in and out are Batch x Time x Dim
        super().__init__(
            f"{in_sig}, skips=None -> {out_sig}",
            Map(Transpose(1, 2), f"{in_sig} -> {in_sig}"),
            *layers,
            (Transpose(1, 2), f"{out_sig} -> {out_sig}"),
            (CausalTrimIfEval(), f"{out_sig} -> {out_sig}")
        )
        self.shift = shift
        self.rf = rf
        self.output_length = (lambda n: n if (self.hp.pad_side != 0) else (n - self.shift + 1))
        self.hp = AttributeDict(init_ctx)

    @staticmethod
    def get_kernels_and_dilation(kernel_sizes, blocks):
        # figure out the layers dilation.
        # User can pass :
        # - 1 kernel, n blocks
        # - 1 block of kernel & n times the length of this block
        # - n kernels for sum(blocks) == n
        # - n kernels & no blocks = 1 block
        if not blocks:
            # single block from kernels
            dilation = accumulate([1, *kernel_sizes], opr.mul)
        else:
            # repetitions of the same block
            if len(set(blocks)) == 1 and set(blocks).pop() == len(kernel_sizes):
                dilation = accumulate([1, *kernel_sizes], opr.mul)
                dilation = chain(*([dilation] * len(blocks)))
                # broadcast!
                kernel_sizes = chain(*([kernel_sizes] * len(blocks)))
            elif len(kernel_sizes) == sum(blocks):
                cum_blocks = list(accumulate(blocks, opr.add))
                dilation = []
                for start, stop in zip([0] + cum_blocks, cum_blocks):
                    ks = kernel_sizes[start:stop-1]
                    dilation += list(accumulate([1, *ks], opr.mul))
            elif len(kernel_sizes) == 1:
                k = kernel_sizes[0]
                # all the same
                kernel_sizes = (k for _ in range(sum(blocks)))
                dilation = (k ** i for block in blocks for i in range(block))
            else:
                raise ValueError(f"number of layers and number of kernel sizes not compatible."
                                 f" Got kernel_sizes={kernel_sizes} ; blocks={blocks}")
        return kernel_sizes, dilation

    def with_io(self, input_modules, output_modules):
        inpt_mod = combine("==", None, *input_modules)
        # inpt_mod = input_modules[0]
        in_pos, in_kw = get_input_signature(inpt_mod.forward)
        self_in = inpt_mod.s.split(" -> ")[1]
        outpt_mod = combine("-<", None, *output_modules)
        out_kwargs = get_input_signature(outpt_mod.forward)[1]
        out_in_vars = ', '.join([v.split("=")[0] for v in out_kwargs.split(',')])
        out_vars = ", ".join([f"y{i}" for i in range(len(output_modules))])
        # initialize the graph without re-initializing the object :
        super().__init__(f"{', '.join([arg for arg in (in_pos, in_kw, out_kwargs) if arg])} -> {out_vars}",
                         (inpt_mod, f"{', '.join([arg for arg in (in_pos, in_kw) if arg])} -> {self_in}"),
                         (self.elevate(), f"{self_in} -> y"),
                         (outpt_mod, f"y, {out_in_vars} -> {out_vars}"))
        return self

    use_fast_generate = False

    def before_generate(self, g_loop, batch, batch_idx):
        if not self.use_fast_generate:
            return

        layers = [m for m in self.modules() if isinstance(m, WNLayer)]

        qs = {i: None for i in range(len(layers))}
        lyr_slices = {}
        dilations = {}
        pads = {}

        for i, layer in enumerate(layers):
            d, k = layer.hp.dilation, layer.hp.kernel_size
            rf = d * k
            if self.hp.pad_side == 0:
                size = (self.rf - layer.hp.cause)
                # indices = [size - n for n in range(0, rf, d)][::-1]
                indices = [*range(-layer.hp.cause - 1, 0, d)]
            else:
                size = (self.rf - 1)
                indices = [size - n for n in range(0, rf, d)][::-1]
            lyr_slices[i] = torch.tensor(indices).long().to(self.device)

        handles = []

        for i, layer in enumerate(layers):

            def pre_hook(mod, inpt, i=i):
                if qs[i] is None:
                    qs[i] = inpt
                    return inpt
                if i == 0 and qs[i] is not None:
                    q = qs[0][:-1]
                    if isinstance(q, tuple):  # multiple inputs!
                        q = [*q]
                        for l in range(len(q)):
                            q[l] = q[l].roll(-1, 2)
                            q[l][:, :, -1:] = inpt[l][:, :, -1:]
                            qs[0] = (*q, None)
                    else:
                        q = q.roll(-1, 2)
                        q[:, :, -1:] = inpt[0][:, :, -1:]
                        qs[0] = (q, None)
                return tuple(x[:, :, lyr_slices[i]]
                             if x is not None else x for x in inpt)

            handles += [layer.register_forward_pre_hook(pre_hook)]

            def hook(module, inpt, outpt, i=i):
                if not getattr(module, '_fast_mode', False):
                    # turn padding and dilation AFTER first pass
                    for mod in module.modules():
                        if isinstance(mod, nn.Conv1d) and mod.dilation != (1,):
                            dilations[mod] = mod.dilation
                            mod.dilation = (1,)
                        if isinstance(mod, CausalPad):
                            pads[mod] = mod.pad
                            mod.pad = (0,) * len(mod.pad)
                    module._fast_mode = True
                    return outpt
                z, skips = outpt[:-1], outpt[-1]
                if i < len(layers) - 1:
                    # set and roll input for next layer
                    q, skp = qs[i + 1][:-1], qs[i + 1][-1]
                    q = [*q]
                    for l in range(len(q)):
                        q[l] = q[l].roll(-1, 2)
                        q[l][:, :, -1:] = z[l][:, :, -1:]
                    if skp is not None:
                        skp = skp.roll(-1, 2)
                        skp[:, :, -1:] = skips[:, :, -1:]

                    qs[i + 1] = (*q, skp)
                return qs.get(i + 1, outpt)

            handles += [layer.register_forward_hook(hook)]
        # save those for later
        return dict(handles=handles, dilations=dilations, pads=pads, layers=layers)

    def generate_step(self, t, inputs, ctx):
        return self.forward(*inputs)

    def after_generate(self, final_outputs, ctx, batch_idx):
        if not self.use_fast_generate:
            return
        # reset the layers' parameters
        for mod, d in ctx['dilations'].items():
            mod.dilation = d
        for mod, pad in ctx['pads'].items():
            mod.pad = pad
        # remove the hooks
        for handle in ctx['handles']:
            handle.remove()
        # turn off fast_mode
        for layer in ctx['layers']:
            layer._fast_mode = False
        return final_outputs



import torch
import numpy as np

from ..modules import angular_distance


def nearest_neighbor(X, Y):
    """
    computes nearest neighbor by angular distance
    """
    D_xy = angular_distance(X, Y)
    dists, nn = torch.min(D_xy, dim=-1)
    return dists, nn


def torch_frame(x, frame_size, hop_length):
    """
    helper to reshape an array into frames
    """
    N = x.size(-1)
    org_size = x.size()[:-1]
    tmp_0 = np.prod(tuple(org_size))
    new_dims = (1 + int((N - frame_size) / hop_length), frame_size)
    framed = torch.as_strided(x.reshape(-1, N), (int(tmp_0), *new_dims), (N, hop_length, 1))
    return framed.reshape(*org_size, *new_dims)


def repeat_rate(x, frame_size, hop_length):
    """
    frames x and compute repeat-rate per frame
    """
    framed = torch_frame(x, frame_size, hop_length)
    uniques = torch.tensor([torch.unique(row).size(0) for row in framed.reshape(-1, framed.size(-1))])
    return (1 - (uniques-1) / (frame_size-1)).reshape(framed.size()[:-1], -1)


# for scoring outputs
def cum_entropy(neighbors, reduce="sum", neg_diff=True):
    """`neighbors` is a (Time, ) Long Tensor [no batch dim!]"""
    items, idx = torch.unique(neighbors, return_inverse=True)
    cum_probs = torch.zeros(items.size(0), neighbors.size(0))
    cum_probs[idx, torch.arange(neighbors.size(0))] = 1
    cum_probs = torch.cumsum(cum_probs, dim=1)
    cum_probs = cum_probs / cum_probs.sum(dim=0, keepdims=True)
    e_wrt_t = (-cum_probs * torch.where(cum_probs > 0, torch.log(cum_probs), cum_probs)).sum(dim=0)
    if neg_diff:
        diff = torch.diff(e_wrt_t, 1, 1, append=torch.tensor([0]).to(e_wrt_t))
        e_wrt_t = (torch.sign(diff) * e_wrt_t)
    return e_wrt_t.sum(dim=0) if reduce == 'sum' else e_wrt_t


def hist_transform(neighbors, bins=256):
    """transform series of neighbors into hist vectors"""
    if neighbors.dim() > 1:
        x_dims = neighbors.shape[:-1]
        h = torch.stack([torch.histc(xi, bins=bins) for xi in neighbors.view(-1, neighbors.size(-1))]).reshape(*x_dims, bins)
        return h
    return torch.histc(neighbors, bins=bins)



from functools import partial
import torch.nn.functional as F
from torch import nn
import torch

from .homs import hom, HOM, Maybe, Sum

__all__ = [
    'Abs',
    'Chunk',
    'Flatten',
    'Transpose',
    'CausalPad',
    'ScaledActivation',
    'ScaledSigmoid',
    'ScaledTanh',
    'ScaledAbs'
]


class Abs(nn.Module):

    def forward(self, x):
        return x.abs_()


class Transpose(nn.Module):

    def __init__(self, dim1, dim2):
        super(Transpose, self).__init__()
        self.dims = (dim1, dim2)

    def forward(self, *args):
        if len(args) > 1:
            return tuple(x.transpose(*self.dims).contiguous() if x is not None else x for x in args)
        return args[0].transpose(*self.dims).contiguous() if args[0] is not None else None


class CausalPad(nn.Module):

    @staticmethod
    def int_to_lr(i):
        return (i, 0) if i >= 0 else (0, abs(i))

    def __init__(self, pad, **kwargs):  # TODO: learn=True
        super().__init__()
        # for whatever reason torch decided dimensions should be in the reversed order...
        self.pad = tuple(i for p in reversed(pad) for i in self.int_to_lr(p))
        self.kwargs = kwargs

    def forward(self, x):
        return F.pad(x, self.pad, **self.kwargs)


class ScaledActivation(nn.Module):
    def __init__(self, activation, dim, with_range=True):
        super(ScaledActivation, self).__init__()
        self.activation = activation
        # self.scales = nn.Parameter(torch.rand(dim, ) * 100, )
        self.scales = nn.Linear(dim, dim)
        self.dim = dim

    def forward(self, x):
        # s = self.scales.to(x).view(*(d if d == self.dim else 1 for d in x.size()))
        # return self.activation(self.rg * x / self.scales) * self.scales
        return self.activation(x) / self.activation(self.scales(x))


class ScaledSigmoid(ScaledActivation):
    def __init__(self, dim, with_range=True):
        super(ScaledSigmoid, self).__init__(nn.Sigmoid(), dim, with_range)


class ScaledTanh(ScaledActivation):
    def __init__(self, dim, with_range=True):
        super(ScaledTanh, self).__init__(nn.Tanh(), dim, with_range)


class ScaledAbs(ScaledActivation):
    def __init__(self, dim, with_range=True):
        super(ScaledAbs, self).__init__(Abs(), dim, with_range)


class Chunk(HOM):
    def __init__(self, mod: nn.Module, chunks, dim=-1, sig_in="x", sum_out=False):
        out_vars = ", ".join(["x" + str(i) for i in range(chunks)])
        super(Chunk, self).__init__(
            f"{sig_in} -> {out_vars if not sum_out else 'out'}",
            (mod, f"{sig_in} -> _tmp_"),
            (partial(torch.chunk, chunks=chunks, dim=dim), f"_tmp_ -> {out_vars}"),
            *Maybe(sum_out,
                   Sum(f"{out_vars} -> out"))
        )


class Flatten(HOM):
    """flatten `n_dims` dimensions of a tensor counting from the last"""

    def __init__(self, n_dims):
        super(Flatten, self).__init__(
            "x -> x",
            (lambda x: x.view(*x.shape[:-n_dims], -1), 'x -> x')
        )



from .parametrized_gaussian import *
from .resamplers import *
from .sample_rnn import *
from .s2s_lstm import *
from .single_class_mlp import *
from .tied_autoencoder import *
from .transformers import *
from .wavenet import *


__all__ = [_ for _ in dir() if not _.startswith("_")]


import torch.nn as nn
import torch

from ..modules.homs import HOM

__all__ = [
    "SingleClassMLP"
]


class SingleClassMLP(nn.Module):

    def __init__(self, d_in, d_mid, d_out,
                 act=nn.Mish(), bias=True,
                 learn_temperature=True, n_hidden_layers=0,
                 net_type="mlp"  # one of ['mlp', 'highway', 'residuals']
                 ):
        super(SingleClassMLP, self).__init__()
        self.learn_temp = learn_temperature
        if net_type == 'mlp':
            self.mlp = nn.Sequential(
                nn.Linear(d_in, d_mid, bias=bias), act,
                *((nn.Linear(d_mid, d_mid, bias=bias), act) * n_hidden_layers),
                nn.Linear(d_mid, d_out + int(learn_temperature), bias=bias)
            )
        elif net_type == 'highway':
            self.mlp = HOM("x -> y",
                           (nn.Sequential(nn.Linear(d_in, d_mid, bias=bias), act, ), "x -> x"),
                           (lambda x, y: torch.cat((x, y), dim=-1), "x, x -> xh"),
                           *(
                                (nn.Sequential(nn.Linear(d_mid * 2, d_mid, bias=bias), act), "xh -> y"),
                                (lambda x, y: torch.cat((x, y), dim=-1), "x, y -> xh"),
                            ) * n_hidden_layers,
                           (nn.Linear(d_mid, d_out + int(learn_temperature), bias=bias), "y -> y"),
                           )
        elif net_type == 'residuals':
            self.mlp = HOM("x -> y",
                           (nn.Sequential(nn.Linear(d_in, d_mid, bias=bias), act, ), "x -> xh"),
                           *(
                                (nn.Sequential(nn.Linear(d_mid, d_mid, bias=bias), act), "xh -> y"),
                                (lambda x, y: x + y, "xh, y -> xh"),
                            ) * n_hidden_layers,
                           (nn.Linear(d_mid, d_out + int(learn_temperature), bias=bias), "xh -> y"),
                           )
        else:
            raise ValueError(f"Expected `net_type` to be one of ['mlp', 'highway', 'residuals']. Got {net_type}")
        self.d_out = d_out

    def forward(self, x, temperature=None):
        outpt = self.mlp(x)
        if self.learn_temp:
            # last output feature is the learned temperature
            outpt = outpt[..., :-1] / (nn.Sigmoid()(outpt[..., -1:]))
        if self.training:
            return outpt
        if temperature is None:
            return outpt.argmax(dim=-1)
        else:
            if not isinstance(temperature, torch.Tensor):
                temperature = torch.Tensor([temperature]).reshape(*([1] * (len(outpt.size()))))
            probas = outpt.squeeze() / temperature.to(outpt)
            probas = nn.Softmax(dim=-1)(probas)
            if probas.dim() > 2:
                o_shape = probas.shape
                probas = probas.view(-1, o_shape[-1])
                return torch.multinomial(probas, 1).reshape(*o_shape[:-1])
            return torch.multinomial(probas, 1)



import numpy as np
import torch
import torch.nn as nn
from librosa.sequence import dtw
from sklearn.metrics import pairwise_distances


__all__ = [
    "optimal_path",
    "NearestNextNeighbor"
]


def optimal_path(x, y):
    return dtw(C=pairwise_distances(abs(x), abs(y), metric='cosine'), subseq=True)[1][::-1]


class NearestNextNeighbor(nn.Module):

    device = property(lambda self: next(self.parameters()).device)

    def __init__(self, feature, snd):
        super(NearestNextNeighbor, self).__init__()
        self.feature = feature
        self.snd = feature.transform(snd[:])
        self._t = -100
        self._starts = None
        self._param = nn.Parameter(torch.ones(1))
        self.shift = self.rf = feature.n_fft
        self.output_length = lambda x: 1

    def predict_start_frame(self, X):
        path = optimal_path(X, self.snd)
        return path[-1, -1]+1

    def generate_step(self, t, inputs, ctx):
        """predict start frame if inputs is new else return next frame"""
        if t != self._t + 1:
            self._starts = [self.predict_start_frame(x.detach().cpu().numpy())
                            for x in inputs[0]]
            self._t = t - 1
        output = np.stack([self.snd[i:i+1] for i in self._starts])
        self._starts = [x+1 for x in self._starts]
        self._t += 1
        return torch.from_numpy(output).to(inputs[0].device)



import numpy as np
from librosa.util import peak_pick, localmax
from librosa.beat import tempo
from librosa import griffinlim
from librosa.sequence import dtw
from scipy.interpolate import RectBivariateSpline as RBS
from scipy.ndimage.filters import minimum_filter1d, median_filter
from sklearn.metrics import pairwise_distances as pwd
from joblib import Parallel, delayed
from typing import List
from numba import njit, prange, float64, intp
import json
import matplotlib.pyplot as plt
import click
import os
import soundfile as sf
import shutil
from time import time
import matplotlib as mpl
mpl.rcParams['agg.path.chunksize'] = 10000

__all__ = [
    'from_recurrence_matrix'
]


def optimal_path(x, y):
    """
    Parameters
    ----------
    x: np.ndarray
        shape: (Time x Dim)
    y: np.ndarray
        shape: (Time x Dim)

    Returns
    -------
    path: np.ndarray
    """
    return dtw(C=pwd(abs(x), abs(y), metric='cosine'), subseq=True)[1][::-1]


def etl(input_file, sr, n_fft, hop_length):
    """STFT Extract-Transform-Load"""
    from mimikit import FileToSignal, MagSpec

    y = FileToSignal(sr)(input_file)
    S = MagSpec(n_fft, hop_length)(y)
    return y, S


@njit(float64[:, :](float64[:, :], intp), fastmath=True, cache=True, parallel=True)
def pwdk_cosine(X, k):
    """
    pairwise distance within a kernel size - cosine version

    Parameters
    ----------
    X: np.ndarray (shape = T x D)
        Matrix of observations in time
    k: int
        kernel size - must be odd and > 0

    Returns
    -------
    diags: np.ndarray (shape = T x k)
        cosine distance between observations t
        and t-(k//2) to t+(k//2)
    """
    T = X.shape[0]
    dist = np.zeros((T, 2 * k - 1))
    kdiv2 = k
    # for angular distance:
    # has_negatives = int(np.any(A < 0))
    for i in prange(X.shape[0]):
        for j in prange(max(i - kdiv2, 0), min(i + kdiv2 + 1, T)):
            if i == j:
                continue
            Xi, Xj = np.ascontiguousarray(X[i]), np.ascontiguousarray(X[j])
            # compute cosine distance
            dij = np.dot(Xi, Xj)
            denom = (np.linalg.norm(Xi) * np.linalg.norm(Xj))
            if denom == 0:
                dij = 1.
            else:
                dij = 1 - (dij / denom)
            # for angular distance:
            # dij = (1 + has_negatives) * np.arccos(dij) / np.pi
            dist[i, kdiv2 + (j - i)] = dij
    return dist


@njit(float64[:](float64[:, :], float64[:, :]), fastmath=True, cache=True, parallel=True)
def convolve_diagonals(diagonals, kernel):
    """
    convolve `diagonals` with `kernel`

    Parameters
    ----------
    diagonals: np.ndarray (T x k)
    kernel: np.ndarray (k x k)

    Returns
    -------
    c: np.ndarray (T, )

    """
    K = kernel.shape[0]
    N = diagonals.shape[0] - K + 1
    out = np.zeros((N,), dtype=np.float64)
    # outer loop is parallel, inner sequential
    for i in prange(N):
        s = 0.
        for j in range(K):
            xi = diagonals[i + j:i + j + 1, K - j - 1:(2 * K) - j - 1]
            xi = np.ascontiguousarray(xi)
            kj = np.ascontiguousarray(kernel[j])
            s = s + np.dot(xi, kj)[0]
        out[i] = s
    return out


def checker(N, normalize=True):
    """
    checker(2)
    =>  [[-1, -1, 0,  1,  1],
         [-1, -1, 0,  1,  1],
         [ 0,  0, 0,  0,  0],
         [ 1,  1, 0, -1, -1],
         [ 1,  1, 0, -1, -1]]
    """
    block = np.zeros((N * 2 + 1, N * 2 + 1), dtype=np.int32)
    for k in range(-N, N + 1):
        for l in range(-N, N + 1):
            block[k + N, l + N] = - np.sign(k) * np.sign(l)
    if normalize:
        block = block / np.abs(block).sum()
    return block.astype(np.float64)


def discontinuity_scores(
        X: np.ndarray,
        kernel_sizes: List[int],
):
    # make the kernels odd
    kernel_sizes = [(k*2)+1 for k in kernel_sizes]
    max_kernel = max(kernel_sizes)
    X = np.ascontiguousarray(X, dtype=np.float)
    N = X.shape[0]
    scores = np.zeros((len(kernel_sizes), N))
    diagonals = pwdk_cosine(X, max_kernel)
    for i, k in enumerate(kernel_sizes):
        kd2 = k // 2
        if k < max_kernel:
            extra = max_kernel - k
            dk = diagonals[:, extra:-extra]
        else:
            dk = diagonals.copy()
        dk = np.pad(dk, ((kd2, kd2), (0, 0)))
        kernel = checker(kd2, normalize=True)
        scr = convolve_diagonals(dk, kernel)
        scores[i] = scr - scr.min()
    return scores


def pick_globally_sorted_maxes(x, wait_before, wait_after, min_strength=0.02):
    mn = minimum_filter1d(
        x, wait_before + wait_after, mode='constant', cval=x.min()
    )
    glob_rg = x.max() - x.min()
    strength = (x - mn) / glob_rg
    # filter out peaks with too few contrasts
    mx = localmax(x) & (strength >= min_strength)

    mx_indices = mx.nonzero()[0][np.argsort(-x[mx])]

    final_maxes = np.zeros_like(x, dtype=bool)

    for m in mx_indices:
        i, j = max(0, m - wait_before), min(x.shape[0], m + wait_after)
        if np.any(final_maxes[i:j]):
            continue
        else:
            # make sure the max dominates left and right
            # aka we are not globally increasing/decreasing around it
            mu_l = x[i:m].mean()
            mu_r = x[m:j].mean()
            mx = x[m]
            if mx > mu_l and mx > mu_r:
                final_maxes[m] = True
    return final_maxes.nonzero()[0]


def from_recurrence_matrix(X,
                           kernel_sizes=[6],
                           min_dur=4,
                           min_strength=0.03,
                           plot=False):
    N = X.shape[0]
    diagonals = discontinuity_scores(X, kernel_sizes)
    dg = diagonals.mean(axis=0)
    mx2 = peak_pick(dg, min_dur // 2, min_dur // 2, min_dur // 2, min_dur // 2, 0., min_dur)
    mx = pick_globally_sorted_maxes(dg, min_dur, min_dur, min_strength)
    mx = mx[(mx > min_dur) & (mx < (N - min_dur))]
    if plot:
        def plot_diagonals():
            plt.figure(figsize=(dg.size // 500, 10))
            for k, d in zip(kernel_sizes, diagonals):
                plt.plot(d, label=f'kernel_size={k}', linestyle='--', alpha=0.75)
            plt.plot(dg, label=f'mean diagonal')
            plt.legend()
            plt.vlines(mx, dg.min(), dg.max(), linestyles='-', alpha=.5, colors='green', label='new method')
            plt.vlines(mx2, dg.min(), dg.max(), linestyles='dotted', alpha=.75, colors='blue', label='old method')
            plt.legend()
            plt.show()
    else:
        def plot_diagonals():
            return
    return mx, plot_diagonals


def export(S, target_path, sr, n_fft, hop_length):
    from mimikit import GLA
    gla = GLA(n_fft, hop_length)
    if S.shape[0] <= 1:
        print("!!!!!", target_path)
        return
    # if S is too long, joblib sends it as np.memmap
    # which messes with gla()...
    y = gla(np.asarray(S))
    sf.write(f"{target_path}.wav", y, sr, 'PCM_24')
    return


@click.command()
@click.option("-f", "--input-file", help="file to be segmented")
@click.option("--sr", "-r", default=22050, help="the sample rate used for loading the file (default=22050)")
@click.option("--n-fft", "-d", default=2048, help="size of the fft (default=2048)")
@click.option("--hop-length", "-h", default=512, help="hop length of the stft (default=512)")
@click.option("--kernel-size", "-k", default=None, type=int,
              help="half size in frames of the kernel "
                   "(if not specified, tempo of the file is estimated and kernel-size is set to 1 beat)")
@click.option("--factor", "-r", multiple=True, type=float,
              help="factor for the kernel size "
                   "(default to 1., can be specified multiple times)")
@click.option("--min-dur", "-m", default=None, type=int,
              help="minimum number of frames per segment "
                   "(if not specified, default to kernel-size)")
@click.option("--min-strength", "-s", default=0.03, type=float,
              help="minimum strength for a peak to be selected "
                   "(default=0.03, must be between 0. and 1.)")
@click.option("--export-durations", "-x", is_flag=True, help="whether to write the durations as a text file")
@click.option("--plot", "-p", is_flag=True, help="whether to plot the results")
def segment(input_file: str,
            sr: int = 22050,
            n_fft=2048,
            hop_length=512,
            kernel_size=None,
            factor=[],
            min_dur=None,
            min_strength=0.03,
            export_durations=False,
            plot=False
            ):
    """extract segments from an audio file"""

    y, S = etl(input_file, sr, n_fft, hop_length)
    if kernel_size is None:
        tmp = tempo(y, sr)[0]
        kernel_size = int((60 / tmp) * sr) // hop_length
    else:
        tmp = None
    if min_dur is None:
        min_dur = kernel_size
    if not factor:
        factor = (1.,)
    start = time()
    stops, plot_diagonals = from_recurrence_matrix(
        S,
        kernel_sizes=[int(f * kernel_size) for f in factor],
        min_dur=min_dur, min_strength=min_strength, plot=plot
    )
    duration = time() - start
    segments = np.split(S, stops, axis=0)
    target_dir = os.path.splitext(input_file)[0]
    durs, counts = np.unique([s.shape[0] for s in segments], return_counts=True)
    distrib_str = "\n    ".join([f"{d}  :  {c}"
                                 for d, c in zip(durs[(-counts).argsort()[:10]],
                                                 counts[(-counts).argsort()[:10]])]) + \
                  "\n    [  ...  ]\n"
    print(
        f"""
    Segmented '{input_file}' in {duration:.3f} seconds

    Original data shape       = {S.shape}
    {f'estimated tempo           = {tmp:.3f} BPM' if tmp is not None else ""}
    1 frame                   = {hop_length / sr:.3f} sec
    kernel size               = {kernel_size} frames ({kernel_size * hop_length / sr:.3f} sec)
    min dur                   = {min_dur} frames ({min_dur * hop_length / sr:.3f} sec)

    number of segments        = {len(segments)}

    MIN  segment duration     = {min(durs)} frames ({np.min(durs) * hop_length / sr:.3f} sec)
    MAX  segment duration     = {max(durs)} frames ({np.max(durs) * hop_length / sr:.3f} sec)
    MEAN segment duration     = {np.mean(durs):.1f} frames ({np.mean(durs) * hop_length / sr:.3f} sec)
    STD  segment duration     = {np.std(durs):.3f} frames ({np.std(durs) * hop_length / sr:.3f} sec)

    Dur : Count
    """ + distrib_str
    )
    print(f"writing segments to target directory '{target_dir}/'")
    if os.path.exists(target_dir):
        shutil.rmtree(target_dir)
    os.makedirs(target_dir, exist_ok=True)
    Parallel(n_jobs=-1, backend='multiprocessing', batch_size=16) \
        (delayed(export)(s, f"{target_dir}/{i}", sr, n_fft, hop_length)
         for i, s in enumerate(segments))

    params = dict(k=kernel_size, r=factor, m=min_dur)
    with open(os.path.join(target_dir, "params.json"), 'w') as f:
        f.write(json.dumps(params))
    if export_durations:
        with open(os.path.join(target_dir, "durations.json"), 'w') as f:
            f.write(json.dumps({i: {"o_dur": s.shape[0], "target": s.shape[0]} for i, s in enumerate(segments)}, indent=1))

    if plot:
        plot_diagonals()
        plt.show()
        plt.figure(figsize=(60, 10))
        plt.plot(y)
        plt.vlines(np.cumsum([s.shape[0] * hop_length for s in segments]), -1, 1, linestyles='--',
                   alpha=.5, colors='green')
        plt.show()
    return


def _stretch_rbs(S, ratio):
    if S.shape[1] <= 1 or S.shape[0] <= 1:
        return S
    time_indices = np.linspace(0, S.shape[1] - 1, int(np.rint(S.shape[1] * ratio)))
    if S.dtype in (np.complex64, np.complex128):
        mag, phase = abs(S), np.imag(S)
    else:
        mag, phase = S, None
    spline = RBS(np.arange(S.shape[0]), np.arange(S.shape[1]), mag)
    interp = spline.ev(np.arange(S.shape[0])[:, None], time_indices)
    if S.dtype in (np.complex64, np.complex128):
        return griffinlim(interp, n_iter=32)
    else:
        return interp


def _stretch(S, target_dur):
    r = target_dur / S.shape[0]
    Shat = _stretch_rbs(S.T, r).T
    return Shat


def nearest_multiple(x, mu):
    return np.maximum(np.round(x / mu) * mu, mu).astype(x.dtype)


def x_pand(x, ga):
    loc, scale = np.mean(x), np.std(x)
    xh = (x - loc) / scale
    y = np.round(loc + (xh * ga) * scale).astype(x.dtype)
    return np.maximum(np.minimum(y, x.max()), 2)


def _kde_stretch(S, target_dur, n_components=32, grid_size=16, smoother_size=2):
    from sklearn.neighbors import KernelDensity
    from sklearn.decomposition import PCA
    from sklearn.model_selection import GridSearchCV

    # project the data to a lower dimension
    pca = PCA(n_components=min(n_components, S.shape[0]), whiten=False)
    data = pca.fit_transform(S)

    # use grid search cross-validation to optimize the bandwidth
    params = {"bandwidth": np.logspace(-1, 1, grid_size)}
    grid = GridSearchCV(KernelDensity(), params)
    grid.fit(data)

    print("best bandwidth: {0}".format(grid.best_estimator_.bandwidth))

    # use the best estimator to compute the kernel density estimate
    kde = grid.best_estimator_

    # sample 44 new points from the data
    new_data = kde.sample(target_dur)
    d = pwd(new_data, data, metric='cosine')
    idx = np.argsort(d.argmin(axis=1))
    new_data = new_data[idx]
    # new_data = median_filter(new_data, size=(smoother_size, smoother_size))
    return pca.inverse_transform(new_data)


@click.command()
@click.option("-s", "--source-dir", help="file to be segmented")
@click.option("--sr", "-r", default=22050, help="the sample rate used for loading the file")
@click.option("--n-fft", "-d", default=2048, help="size of the fft")
@click.option("--hop-length", "-h", default=512, help="hop length of the stft")
@click.option("--nmm", "-n", default=None, type=int,
              help="stretch segments to the nearest multiple of the specified value")
@click.option("--xpand", "-x", default=None, type=float,
              help="expand (>1.) or compress (<1.) durations around their mean")
@click.option("--manual", "-m", is_flag=True,
              help="if specified, expects `source-dir` to contain a `durations.json` file "
                   "containing target durations for each segment")
@click.option("--stretcher", "-t", type=str, help="method to use for stretching. One of 'rbs', 'kde'",
              default='rbs')
@click.option("--verbose", "-v", is_flag=True,
              help="whether to print out the transformation durations")
def re_stretch(source_dir: str,
               sr=22050,
               n_fft=2048,
               hop_length=512,
               nmm=None,
               xpand=None,
               manual=False,
               stretcher='rbs',
               verbose=False):
    from h5mapper import FileWalker

    files = sorted(FileWalker(r'(.*/[0-9]*\.wav$)', source_dir),
                   key=lambda f: int(os.path.splitext(os.path.split(f)[1])[0]))
    segments = Parallel(n_jobs=-1, backend="multiprocessing")(
        delayed(etl)(f, sr, n_fft, hop_length)
        for f in files
    )
    segments = [s[1] for s in segments]
    durations = np.r_[[s.shape[0] for s in segments]]
    with open(os.path.join(source_dir, "params.json"), 'r') as f:
        arg_str = "_".join([k+str(v) for k, v in json.loads(f.read()).items()])
    if nmm is not None:
        targets = nearest_multiple(durations, nmm)
        arg_str += f"_n{nmm}"
    elif xpand is not None:
        targets = x_pand(durations, xpand)
        arg_str += f"_x{xpand}"
    elif manual and os.path.isfile(os.path.join(source_dir, "durations.json")):
        dct = json.load(open(os.path.join(source_dir, "durations.json"), 'r'))
        targets = np.zeros((durations.size,), dtype=np.int)
        for i, data in dct.items():
            targets[int(i)] = data['target']
        arg_str += f"_manual"
    else:
        raise ValueError("`nmm` and `xpand` are None, `manual` is False. Cannot stretch without targets.")
    if stretcher == 'rbs':
        stretch_func = _stretch
    else:
        stretch_func = _kde_stretch
    if verbose:
        for i, (o, t) in enumerate(zip(durations, targets)):
            print(f"{i}: {o} -> {t}")
    stretched = Parallel(n_jobs=-1, backend="multiprocessing")(
        delayed(stretch_func)(s, target)
        for s, target in zip(segments, targets)
    )
    stretched = np.concatenate(stretched, axis=0)
    export(stretched, os.path.join(source_dir, f"stretched_{arg_str}"), sr, n_fft, hop_length)
    return
